#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

#
# Note pickling is used for message serialization.
# This is just for the prototyping & fast PoC, pickling wont be used in the production.
# Instead, protobuf messages will be defined and parsed to avoid malicious pickling.
#

import logging
import os

import requests
from requests.auth import HTTPDigestAuth

logger = logging.getLogger(__name__)


class WalletRpc(object):
    """
    RPC helper
    """

    def __init__(self, agent, port=None, creds=None):
        self.agent = agent
        self.port = port
        self.creds = creds
        self.url = None
        self.set_addr("127.0.0.1:%s" % port)

    def set_addr(self, addr):
        self.url = "http://%s/json_rpc" % addr

    def set_creds(self, creds):
        if creds is None or (isinstance(creds, (list, tuple)) and len(creds) == 2):
            self.creds = creds
        elif isinstance(creds, str):
            self.creds = creds.split(":", 1)
        else:
            raise ValueError("Unknown creds type")

    def request(self, method, params=None):
        """
        Request wrapper
        {"jsonrpc":"2.0","id":"0","method":"get_address", "params":{"account_index": 0, "address_index": [0,1,2,3,4,5]}
        :param method:
        :param params:
        :return:
        """
        auth = HTTPDigestAuth(self.creds[0], self.creds[1]) if self.creds else None
        js = {"jsonrpc": "2.0", "id": "0", "method": method}
        if params:
            js["params"] = params

        resp = requests.post(self.url, json=js, auth=auth)
        resp.raise_for_status()
        return resp.json()

    def balance(self):
        return self.request("getbalance")

    def height(self):
        return self.request("getheight")

    def get_transfers(self, params=None):
        return self.request("get_transfers", params)

    def rescan_bc(self):
        return self.request("rescan_blockchain")

    def transfer(self, params):
        return self.request("transfer", params)

    def submit_transfer(self, params):
        return self.request("submit_transfer", params)

    def stop_wallet(self):
        return self.request("stop_wallet")

    def export_outputs(self):
        return self.request("export_outputs")

    def import_outputs(self, params=None):
        return self.request("import_outputs", params)

    def import_key_images(self, params=None):
        return self.request("import_key_images", params)

    def refresh(self, params=None):
        return self.request("refresh", params)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import binascii
import struct

from monero_glue.misc import b58_mnr
from monero_glue.xmr import common, crypto, mlsag2, ring_ct
from monero_serialize import protobuf as xproto
from monero_serialize import xmrserialize, xmrtypes
from .sub.addr import *
from .sub.creds import *
from .sub.keccak_hasher import *
from .sub.mlsag_hasher import *
from .sub.recode import *
from .sub.recode_ext import *
from .sub.tsx_helper import *
from .sub.xmr_net import *


DISPLAY_DECIMAL_POINT = 12


class XmrNoSuchAddressException(common.XmrException):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TxScanInfo(object):
    """
    struct tx_scan_info_t
    """

    __slots__ = [
        "in_ephemeral",
        "ki",
        "mask",
        "amount",
        "money_transfered",
        "error",
        "received",
    ]


def get_subaddress_secret_key(
    secret_key, index=None, major=None, minor=None, little_endian=True
):
    """
    Builds subaddress secret key from the subaddress index
    Hs(SubAddr || a || index_major || index_minor)

    UPDATE: Monero team fixed this problem. Always use little endian.
    Note: need to handle endianity in the index
    C-code simply does: memcpy(data + sizeof(prefix) + sizeof(crypto::secret_key), &index, sizeof(subaddress_index));
    Where the index has the following form:

    struct subaddress_index {
        uint32_t major;
        uint32_t minor;
    }

    https://docs.python.org/3/library/struct.html#byte-order-size-and-alignment
    :param secret_key:
    :param index:
    :param major:
    :param minor:
    :param little_endian:
    :return:
    """
    if index:
        major = index.major
        minor = index.minor
    endianity = "<" if little_endian else ">"
    prefix = b"SubAddr"
    buffer = bytearray(len(prefix) + 1 + 32 + 4 + 4)
    struct.pack_into(
        "%s7sb32sLL" % endianity,
        buffer,
        0,
        prefix,
        0,
        crypto.encodeint(secret_key),
        major,
        minor,
    )
    return crypto.hash_to_scalar(buffer)


def get_subaddress_spend_public_key(view_private, spend_public, major, minor):
    """
    Generates subaddress spend public key D_{major, minor}
    :param view_private:
    :param spend_public:
    :param major:
    :param minor:
    :return:
    """
    m = get_subaddress_secret_key(view_private, major=major, minor=minor)
    M = crypto.scalarmult_base(m)
    D = crypto.point_add(spend_public, M)
    return D


def generate_key_derivation(pub_key, priv_key):
    """
    Generates derivation priv_key * pub_key.
    Simple ECDH.
    :param pub_key:
    :param priv_key:
    :return:
    """
    return crypto.generate_key_derivation(pub_key, priv_key)


def derive_subaddress_public_key(out_key, derivation, output_index):
    """
    out_key - H_s(derivation || varint(output_index))G
    :param out_key:
    :param derivation:
    :param output_index:
    :return:
    """
    crypto.check_ed25519point(out_key)
    scalar = crypto.derivation_to_scalar(derivation, output_index)
    point2 = crypto.scalarmult_base(scalar)
    point4 = crypto.point_sub(out_key, point2)
    return point4


def generate_key_image(public_key, secret_key):
    """
    Key image: secret_key * H_p(pub_key)
    :param public_key: encoded point
    :param secret_key:
    :return:
    """
    point = crypto.hash_to_ec(public_key)
    point2 = crypto.ge_scalarmult(secret_key, point)
    return point2


def is_out_to_acc_precomp(
    subaddresses, out_key, derivation, additional_derivations, output_index
):
    """
    Searches subaddresses for the computed subaddress_spendkey.
    If found, returns (major, minor), derivation.

    :param subaddresses:
    :param out_key:
    :param derivation:
    :param additional_derivations:
    :param output_index:
    :return:
    """
    subaddress_spendkey = crypto.encodepoint(
        derive_subaddress_public_key(out_key, derivation, output_index)
    )
    if subaddress_spendkey in subaddresses:
        return subaddresses[subaddress_spendkey], derivation

    if additional_derivations and len(additional_derivations) > 0:
        if output_index >= len(additional_derivations):
            raise ValueError("Wrong number of additional derivations")

        subaddress_spendkey = derive_subaddress_public_key(
            out_key, additional_derivations[output_index], output_index
        )
        subaddress_spendkey = crypto.encodepoint(subaddress_spendkey)
        if subaddress_spendkey in subaddresses:
            return (
                subaddresses[subaddress_spendkey],
                additional_derivations[output_index],
            )

    return None


def generate_key_image_helper_precomp(
    ack, out_key, recv_derivation, real_output_index, received_index
):
    """
    Generates UTXO spending key and key image.

    :param ack: sender credentials
    :type ack: AccountCreds
    :param out_key: real output (from input RCT) destination key
    :param recv_derivation:
    :param real_output_index:
    :param received_index: subaddress index this payment was received to
    :return:
    """
    if ack.spend_key_private == 0:
        raise ValueError("Watch-only wallet not supported")

    # derive secret key with subaddress - step 1: original CN derivation
    scalar_step1 = crypto.derive_secret_key(
        recv_derivation, real_output_index, ack.spend_key_private
    )

    # step 2: add Hs(SubAddr || a || index_major || index_minor)
    subaddr_sk = None
    scalar_step2 = None
    if received_index == (0, 0):
        scalar_step2 = scalar_step1
    else:
        subaddr_sk = get_subaddress_secret_key(
            ack.view_key_private, major=received_index[0], minor=received_index[1]
        )
        scalar_step2 = crypto.sc_add(scalar_step1, subaddr_sk)

    # when not in multisig, we know the full spend secret key, so the output pubkey can be obtained by scalarmultBase
    if len(ack.multisig_keys) == 0:
        pub_ver = crypto.scalarmult_base(scalar_step2)

    else:
        # When in multisig, we only know the partial spend secret key. But we do know the full spend public key,
        # so the output pubkey can be obtained by using the standard CN key derivation.
        pub_ver = crypto.derive_public_key(
            recv_derivation, real_output_index, ack.spend_key_public
        )

        # Add the contribution from the subaddress part
        if received_index != (0, 0):
            subaddr_pk = crypto.scalarmult_base(subaddr_sk)
            pub_ver = crypto.point_add(pub_ver, subaddr_pk)

    if not crypto.point_eq(pub_ver, out_key):
        raise ValueError(
            "key image helper precomp: given output pubkey doesn't match the derived one"
        )

    ki = generate_key_image(crypto.encodepoint(pub_ver), scalar_step2)
    return scalar_step2, ki


def generate_key_image_helper(
    creds,
    subaddresses,
    out_key,
    tx_public_key,
    additional_tx_public_keys,
    real_output_index,
):
    """
    Generates UTXO spending key and key image.
    Supports subaddresses.

    :param creds:
    :param subaddresses:
    :param out_key: real output (from input RCT) destination key
    :param tx_public_key: real output (from input RCT) public key
    :param additional_tx_public_keys:
    :param real_output_index: index of the real output in the RCT
    :return:
    """
    recv_derivation = generate_key_derivation(tx_public_key, creds.view_key_private)

    additional_recv_derivations = []
    for add_pub_key in additional_tx_public_keys:
        additional_recv_derivations.append(
            generate_key_derivation(add_pub_key, creds.view_key_private)
        )

    subaddr_recv_info = is_out_to_acc_precomp(
        subaddresses,
        out_key,
        recv_derivation,
        additional_recv_derivations,
        real_output_index,
    )
    if subaddr_recv_info is None:
        raise XmrNoSuchAddressException()

    xi, ki = generate_key_image_helper_precomp(
        creds, out_key, subaddr_recv_info[1], real_output_index, subaddr_recv_info[0]
    )
    return xi, ki, recv_derivation


def check_acc_out_precomp(tx_out, subaddresses, derivation, additional_derivations, i):
    """
    wallet2::check_acc_out_precomp
    Detects whether the tx output belongs to the subaddresses. If yes, computes the derivation.
    Returns TxScanInfo

    :param tx_out:
    :param derivation:
    :param additional_derivations:
    :param i:
    :return:
    """
    tx_scan_info = TxScanInfo()
    tx_scan_info.error = True

    if not isinstance(tx_out.target, xmrtypes.TxoutToKey):
        return tx_scan_info

    tx_scan_info.received = is_out_to_acc_precomp(
        subaddresses,
        crypto.decodepoint(tx_out.target.key),
        derivation,
        additional_derivations,
        i,
    )
    if tx_scan_info.received:
        tx_scan_info.money_transfered = tx_out.amount
    else:
        tx_scan_info.money_transfered = 0
    tx_scan_info.error = False
    return tx_scan_info


def scan_output(creds, tx, i, tx_scan_info, tx_money_got_in_outs, outs, multisig):
    """
    Wallet2::scan_output()
    Computes spending key, key image, decodes ECDH info, amount, checks masks.

    :param creds:
    :param tx:
    :param i:
    :param tx_scan_info:
    :param tx_money_got_in_outs:
    :param outs:
    :param multisig:
    :return:
    """
    if multisig:
        tx_scan_info.in_ephemeral = 0
        tx_scan_info.ki = crypto.identity()

    else:
        out_dec = crypto.decodepoint(tx.vout[i].target.key)
        res = generate_key_image_helper_precomp(
            creds, out_dec, tx_scan_info.received[1], i, tx_scan_info.received[0]
        )
        tx_scan_info.in_ephemeral, tx_scan_info.ki = res
        if not tx_scan_info.ki:
            raise ValueError("Key error generation failed")

    outs.append(i)
    if tx_scan_info.money_transfered == 0:
        res2 = ecdh_decode_rv(tx.rct_signatures, tx_scan_info.received[1], i)
        tx_scan_info.money_transfered, tx_scan_info.mask = res2
        tx_scan_info.money_transfered = crypto.sc_get64(tx_scan_info.money_transfered)

    tx_money_got_in_outs[tx_scan_info.received[0]] += tx_scan_info.money_transfered
    tx_scan_info.amount = tx_scan_info.money_transfered
    return tx_scan_info


def ecdh_decode_rv(rv, derivation, i):
    """
    Decodes ECDH info from transaction.

    :param rv:
    :param derivation:
    :param i:
    :return:
    """
    scalar = crypto.derivation_to_scalar(derivation, i)
    if rv.type in [xmrtypes.RctType.Simple, xmrtypes.RctType.SimpleBulletproof]:
        return ecdh_decode_simple(rv, scalar, i)

    elif rv.type in [xmrtypes.RctType.Full, xmrtypes.RctType.FullBulletproof]:
        return ecdh_decode_simple(rv, scalar, i)

    else:
        raise ValueError("Unknown rv type")


def ecdh_decode_simple(rv, sk, i):
    """
    Decodes ECDH from the transaction, checks mask (decoding validity).

    :param rv:
    :param sk:
    :param i:
    :return:
    """
    if i >= len(rv.ecdhInfo):
        raise ValueError("Bad index")
    if len(rv.outPk) != len(rv.ecdhInfo):
        raise ValueError("outPk vs ecdhInfo mismatch")

    ecdh_info = rv.ecdhInfo[i]
    ecdh_info = recode_ecdh(ecdh_info, False)
    ecdh_info = ring_ct.ecdh_decode(ecdh_info, derivation=crypto.encodeint(sk))
    c_tmp = crypto.add_keys2(ecdh_info.mask, ecdh_info.amount, crypto.gen_H())
    if not crypto.point_eq(c_tmp, crypto.decodepoint(rv.outPk[i].mask)):
        raise ValueError("Amount decoded incorrectly")

    return ecdh_info.amount, ecdh_info.mask


async def get_transaction_prefix_hash(tx):
    """
    Computes transaction prefix in one step
    :param tx:
    :return:
    """
    writer = get_keccak_writer()
    ar1 = xmrserialize.Archive(writer, True)
    await ar1.message(tx, msg_type=xmrtypes.TransactionPrefixExtraBlob)
    return writer.get_digest()


def expand_transaction(tx):
    """
    Expands transaction - recomputes fields not serialized.
    Recomputes only II, does not have access to the blockchain to get public keys for inputs
    for mix ring reconstruction.

    :param tx:
    :return:
    """
    rv = tx.rct_signatures
    if rv.type in [xmrtypes.RctType.Full, xmrtypes.RctType.FullBulletproof]:
        rv.p.MGs[0].II = [None] * len(tx.vin)
        for n in range(len(tx.vin)):
            rv.p.MGs[0].II[n] = tx.vin[n].k_image

    elif rv.type in [xmrtypes.RctType.Simple, xmrtypes.RctType.SimpleBulletproof]:
        if len(rv.p.MGs) != len(tx.vin):
            raise ValueError("Bad MGs size")
        for n in range(len(tx.vin)):
            rv.p.MGs[n].II = [tx.vin[n].k_image]

    else:
        raise ValueError("Unsupported rct tx type %s" % rv.type)

    return tx


def compute_subaddresses(creds, account, indices, subaddresses=None):
    """
    Computes subaddress public spend key for receiving transactions.

    :param creds: credentials
    :param account: major index
    :param indices: array of minor indices
    :param subaddresses: subaddress dict. optional.
    :return:
    """
    if subaddresses is None:
        subaddresses = {}

    for idx in indices:
        if account == 0 and idx == 0:
            subaddresses[crypto.encodepoint(creds.spend_key_public)] = (0, 0)
            continue

        pub = get_subaddress_spend_public_key(
            creds.view_key_private, creds.spend_key_public, major=account, minor=idx
        )
        pub = crypto.encodepoint(pub)
        subaddresses[pub] = (account, idx)
    return subaddresses


async def get_tx_pub_key_from_received_outs(td):
    """
    Extracts tx pub key from extras.
    Handles previous bug in Monero.

    :param td:
    :type td: xmrtypes.TransferDetails
    :return:
    """
    extras = await parse_extra_fields(list(td.m_tx.extra))
    tx_pub = find_tx_extra_field_by_type(extras, xmrtypes.TxExtraPubKey, 0)

    # Due to a previous bug, there might be more than one tx pubkey in extra, one being
    # the result of a previously discarded signature.
    # For speed, since scanning for outputs is a slow process, we check whether extra
    # contains more than one pubkey. If not, the first one is returned. If yes, they're
    # checked for whether they yield at least one output
    second_pub = find_tx_extra_field_by_type(extras, xmrtypes.TxExtraPubKey, 1)
    if second_pub is None:
        return tx_pub.pub_key

    # Workaround: resend all your funds to the wallet in a different transaction.
    # Proper handling would require derivation -> need trezor roundtrips.
    raise ValueError("Input transaction is buggy, contains two tx keys")


def generate_keys(recovery_key):
    """
    Wallet gen.
    :param recovery_key:
    :return:
    """
    sec = crypto.sc_reduce32(recovery_key)
    pub = crypto.scalarmult_base(sec)
    return sec, pub


def generate_monero_keys(seed):
    """
    Generates spend key / view key from the seed in the same manner as Monero code does.

    account.cpp:
    crypto::secret_key account_base::generate(const crypto::secret_key& recovery_key, bool recover, bool two_random).
    :param seed:
    :return:
    """
    spend_sec, spend_pub = generate_keys(crypto.decodeint(seed))
    hash = crypto.cn_fast_hash(crypto.encodeint(spend_sec))
    view_sec, view_pub = generate_keys(crypto.decodeint(hash))
    return spend_sec, spend_pub, view_sec, view_pub


def generate_sub_address_keys(view_sec, spend_pub, major, minor):
    m = get_subaddress_secret_key(view_sec, major=major, minor=minor)
    M = crypto.scalarmult_base(m)
    D = crypto.point_add(spend_pub, M)
    C = crypto.ge_scalarmult(view_sec, D)
    return D, C



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class NEMImportanceTransfer(p.MessageType):
    FIELDS = {
        1: ('mode', p.UVarintType, 0),
        2: ('public_key', p.BytesType, 0),
    }

    def __init__(
        self,
        mode: int = None,
        public_key: bytes = None,
    ) -> None:
        self.mode = mode
        self.public_key = public_key



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class LiskAddress(p.MessageType):
    MESSAGE_WIRE_TYPE = 115
    FIELDS = {
        1: ('address', p.UnicodeType, 0),
    }

    def __init__(
        self,
        address: str = None,
    ) -> None:
        self.address = address



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class SetU2FCounter(p.MessageType):
    MESSAGE_WIRE_TYPE = 63
    FIELDS = {
        1: ('u2f_counter', p.UVarintType, 0),
    }

    def __init__(
        self,
        u2f_counter: int = None,
    ) -> None:
        self.u2f_counter = u2f_counter



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import collections

from monero_glue.messages import (
    MoneroExportedKeyImage,
    MoneroKeyImageExportInitRequest,
    MoneroSubAddressIndicesList,
    MoneroTransferDetails,
)
from monero_glue.xmr import common, crypto, mlsag2, monero, ring_ct
from monero_serialize import xmrserialize, xmrtypes


async def yield_key_image_data(outputs):
    """
    Process outputs, yields out_key, tx pub key, additional tx pub keys data
    yield in async from py3.6

    :param outputs:
    :return:
    """
    res = []
    for idx, td in enumerate(outputs):  # type: xmrtypes.TransferDetails
        if common.is_empty(td.m_tx.vout):
            raise ValueError("Tx with no outputs %s" % idx)

        tx_pub_key = await monero.get_tx_pub_key_from_received_outs(td)
        extras = await monero.parse_extra_fields(list(td.m_tx.extra))
        additional_pub_keys = monero.find_tx_extra_field_by_type(
            extras, xmrtypes.TxExtraAdditionalPubKeys
        )
        out_key = td.m_tx.vout[td.m_internal_output_index].target.key
        cres = MoneroTransferDetails(
            out_key=out_key,
            tx_pub_key=tx_pub_key,
            additional_tx_pub_keys=additional_pub_keys.data
            if additional_pub_keys
            else None,
            internal_output_index=td.m_internal_output_index,
        )
        res.append(cres)
    return res


def compute_hash(rr):
    """
    Hash over output to ki-sync
    :param rr:
    :type rr: TransferDetails
    :return:
    """
    buff = b""
    buff += rr.out_key
    buff += rr.tx_pub_key
    if rr.additional_tx_pub_keys:
        buff += b"".join(rr.additional_tx_pub_keys)
    buff += xmrserialize.dump_uvarint_b(rr.internal_output_index)

    return crypto.cn_fast_hash(buff)


async def generate_commitment(outputs):
    """
    Generates num, hash commitment for initial message for ki syc
    :param outputs:
    :type outputs: list[xmrtypes.TransferDetails]
    :return:
    """
    hashes = []
    sub_indices = collections.defaultdict(lambda: set())
    for out in outputs:
        sub_indices[out.m_subaddr_index.major].add(out.m_subaddr_index.minor)

    num = 0
    iter = await yield_key_image_data(outputs)
    for rr in iter:  # type: MoneroTransferDetails
        hash = compute_hash(rr)
        hashes.append(hash)
        num += 1

    final_hash = crypto.cn_fast_hash(b"".join(hashes))
    indices = []

    for major in sub_indices:
        indices.append(
            MoneroSubAddressIndicesList(
                account=major, minor_indices=list(sub_indices[major])
            )
        )

    return MoneroKeyImageExportInitRequest(num=num, hash=final_hash, subs=indices)


async def export_key_image(creds, subaddresses, td):
    """
    Key image export
    :param creds:
    :param subaddresses:
    :param td:
    :return:
    """
    out_key = crypto.decodepoint(td.out_key)
    tx_pub_key = crypto.decodepoint(td.tx_pub_key)
    additional_tx_pub_keys = []
    if not common.is_empty(td.additional_tx_pub_keys):
        additional_tx_pub_keys = [
            crypto.decodepoint(x) for x in td.additional_tx_pub_keys
        ]

    ki, sig = ring_ct.export_key_image(
        creds,
        subaddresses,
        out_key,
        tx_pub_key,
        additional_tx_pub_keys,
        td.internal_output_index
        if isinstance(td, MoneroTransferDetails)
        else td.m_internal_output_index,
    )

    return ki, sig



from Crypto.Protocol.KDF import PBKDF2
import hmac
import hashlib


def mnemonics_to_seed(seed, passphrase=b""):
    salt = b"mnemonic" + passphrase

    def prf(p, s):
        hx = hmac.new(p, msg=s, digestmod=hashlib.sha512)
        return hx.digest()

    res = PBKDF2(password=seed, salt=salt, dkLen=64, prf=prf, count=2048)
    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018


class TokenInterface(object):
    def __init__(self, ctx=None):
        self.ctx = ctx

    def gctx(self, ctx):
        return ctx if not None else self.ctx

    async def confirm_transaction(self, tsx_data, creds=None, ctx=None):
        """
        Ask for confirmation from user
        :param tsx_data:
        :param creds:
        :param ctx:
        :return:
        """
        return True

    async def transaction_error(self, *args, **kwargs):
        """
        Transaction error
        :return:
        """

    async def transaction_signed(self, ctx=None):
        """
        Notifies the transaction was completely signed
        :return:
        """

    async def transaction_finished(self, ctx=None):
        """
        Notifies the transaction has been completed (all data were sent)
        :return:
        """

    async def transaction_step(self, step, sub_step=None, sub_step_total=None):
        """
        Transaction progress
        :param step:
        :param sub_step:
        :param sub_step_total:
        :return:
        """

    async def confirm_ki_sync(self, init_msg, ctx=None):
        """
        Ask confirmation on key image sync
        :param init_msg:
        :return:
        """
        return True

    async def ki_error(self, e, ctx=None):
        """
        Key image sync error
        :param e:
        :return:
        """

    async def ki_step(self, i, ctx=None):
        """
        Key image sync step
        :param i:
        :return:
        """

    async def ki_finished(self, ctx=None):
        """
        Ki sync finished
        :return:
        """


def get_iface(ctx=None):
    return TokenInterface(ctx)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import binascii
import os
import unittest
import zlib

import pkg_resources
from monero_serialize import xmrboost, xmrserialize, xmrtypes

from monero_glue.agent import agent_lite
from monero_glue.hwtoken import token
from monero_glue.xmr import crypto, monero, wallet
from monero_glue.xmr.sub.seed import SeedDerivation
from monero_glue_test.base_agent_test import BaseAgentTest
from monero_glue.trezor import manager as tmanager


class TrezorTest(BaseAgentTest):
    def __init__(self, *args, **kwargs):
        super(TrezorTest, self).__init__(*args, **kwargs)
        self.trezor_proxy = None  # type: tmanager.Trezor
        self.agent = None  # type: agent_lite.Agent
        self.creds = None

    def get_trezor_path(self):
        tpath = os.getenv('TREZOR_PATH')
        return tpath if tpath is not None else 'udp:127.0.0.1:21324'

    def reinit_trezor(self):
        self.deinit()
        path = self.get_trezor_path()
        self.creds = self.get_trezor_creds(0)
        self.trezor_proxy = tmanager.Trezor(path=path, debug=True)
        self.agent = agent_lite.Agent(self.trezor_proxy, network_type=monero.NetworkTypes.TESTNET)

        client = self.trezor_proxy.client
        client.transport.session_begin()
        client.wipe_device()
        client.load_device_by_mnemonic(
            mnemonic=self.get_trezor_mnemonics()[0],
            pin="",
            passphrase_protection=False,
            label="ph4test",
            language="english",
        )
        client.transport.session_end()

    def deinit(self):
        try:
            if self.trezor_proxy and self.trezor_proxy.client:
                self.trezor_proxy.client.close()
        except Exception as e:
            pass

    def setUp(self):
        super().setUp()
        self.reinit_trezor()

    def tearDown(self):
        self.deinit()
        super().tearDown()

    async def test_ping(self):
        await self.trezor_proxy.ping()

    async def test_get_address(self):
        res = await self.agent.get_address()
        self.assertIsNotNone(res)
        self.assertEqual(res.address, self.creds.address)

    async def test_get_watch(self):
        res = await self.agent.get_watch_only()
        self.assertIsNotNone(res)
        self.assertEqual(res.watch_key, crypto.encodeint(self.creds.view_key_private))
        self.assertEqual(res.address, self.creds.address)

    async def test_ki_sync(self):
        ki_data = self.get_data_file("ki_sync_01.txt")
        ki_loaded = await wallet.load_exported_outputs(
            self.creds.view_key_private, ki_data
        )

        self.assertEqual(ki_loaded.m_spend_public_key, crypto.encodepoint(self.creds.spend_key_public))
        self.assertEqual(ki_loaded.m_view_public_key, crypto.encodepoint(self.creds.view_key_public))
        res = await self.agent.import_outputs(ki_loaded.tds)
        await self.verify_ki_export(res, ki_loaded)

    async def test_transactions(self):
        files = self.get_trezor_tsx_tests()
        creds = self.get_trezor_creds(0)
        all_creds = [self.get_trezor_creds(0), self.get_trezor_creds(1), self.get_trezor_creds(2)]

        for fl in files:
            with self.subTest(msg=fl):
                unsigned_tx_c = self.get_data_file(fl)
                unsigned_tx = await wallet.load_unsigned_tx(
                    creds.view_key_private, unsigned_tx_c
                )

                await self.tx_sign_test(self.agent, unsigned_tx, creds, all_creds, fl)



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class MoneroTransactionSetOutputAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 506
    FIELDS = {
        1: ('tx_out', p.BytesType, 0),
        2: ('vouti_hmac', p.BytesType, 0),
        3: ('rsig', p.BytesType, 0),
        4: ('out_pk', p.BytesType, 0),
        5: ('ecdh_info', p.BytesType, 0),
    }

    def __init__(
        self,
        tx_out: bytes = None,
        vouti_hmac: bytes = None,
        rsig: bytes = None,
        out_pk: bytes = None,
        ecdh_info: bytes = None,
    ) -> None:
        self.tx_out = tx_out
        self.vouti_hmac = vouti_hmac
        self.rsig = rsig
        self.out_pk = out_pk
        self.ecdh_info = ecdh_info



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class TezosSignedTx(p.MessageType):
    MESSAGE_WIRE_TYPE = 153
    FIELDS = {
        1: ('signature', p.BytesType, 0),
        2: ('sig_op_contents', p.BytesType, 0),
        3: ('operation_hash', p.UnicodeType, 0),
    }

    def __init__(
        self,
        signature: bytes = None,
        sig_op_contents: bytes = None,
        operation_hash: str = None,
    ) -> None:
        self.signature = signature
        self.sig_op_contents = sig_op_contents
        self.operation_hash = operation_hash



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
from .StellarAssetType import StellarAssetType


class StellarChangeTrustOp(p.MessageType):
    MESSAGE_WIRE_TYPE = 216
    FIELDS = {
        1: ('source_account', p.UnicodeType, 0),
        2: ('asset', StellarAssetType, 0),
        3: ('limit', p.UVarintType, 0),
    }

    def __init__(
        self,
        source_account: str = None,
        asset: StellarAssetType = None,
        limit: int = None,
    ) -> None:
        self.source_account = source_account
        self.asset = asset
        self.limit = limit



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import binascii
import os
import unittest
import zlib

from monero_serialize import xmrboost, xmrserialize, xmrtypes

from monero_glue.agent import agent_lite
from monero_glue.hwtoken import token
from monero_glue.xmr import crypto, monero, wallet
from monero_glue_test.base_agent_test import BaseAgentTest


class AgentLiteTest(BaseAgentTest):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(AgentLiteTest, self).__init__(*args, **kwargs)

    async def test_tx_sign_simple(self):
        """
        Testing tx signature, simple, multiple inputs
        :return:
        """
        unsigned_tx_c = self.get_data_file("tsx_uns01.txt")
        unsigned_tx = zlib.decompress(binascii.unhexlify(unsigned_tx_c))
        await self.tx_sign_unsigned(unsigned_tx, "tsx_uns01.txt")

    async def test_tx_sign(self):
        """
        Testing tx signature, one input. non-simple RCT
        :return:
        """
        unsigned_tx_c = self.get_data_file("tsx_uns02.txt")
        unsigned_tx = zlib.decompress(binascii.unhexlify(unsigned_tx_c))
        await self.tx_sign_unsigned(unsigned_tx, "tsx_uns02.txt")

    async def test_tx_sign_sub_dest(self):
        """
        Testing tx signature, one input. non-simple RCT
        :return:
        """
        unsigned_tx_c = self.get_data_file("tsx_uns03.txt")
        unsigned_tx = zlib.decompress(binascii.unhexlify(unsigned_tx_c))
        await self.tx_sign_unsigned(unsigned_tx, "tsx_uns03.txt")

    async def test_tx_sign_sub_2dest(self):
        """
        Testing tx signature, one input. non-simple RCT
        :return:
        """
        unsigned_tx_c = self.get_data_file("tsx_uns04.txt")
        unsigned_tx = zlib.decompress(binascii.unhexlify(unsigned_tx_c))
        await self.tx_sign_unsigned(unsigned_tx, "tsx_uns04.txt")

    async def test_tx_sign_pending01_boost_full_2dest(self):
        """
        Testing tx signature, one input. full RCT, boost serialized
        :return:
        """
        pending_hex = self.get_data_file("tsx_pending01.txt")
        pending_bin = binascii.unhexlify(pending_hex)
        await self.tx_sign_pending_boost(pending_bin, "tsx_pending01.txt")

    async def test_tx_sign_pending02_boost_full_sub_3dest(self):
        """
        Testing tx signature, one input. full RCT, boost serialized
        :return:
        """
        pending_hex = self.get_data_file("tsx_pending02.txt")
        pending_bin = binascii.unhexlify(pending_hex)
        await self.tx_sign_pending_boost(pending_bin, "tsx_pending02.txt")

    async def test_tx_sign_uns01_boost_full_2dest(self):
        """
        Testing tx signature, one input. full RCT, boost serialized
        :return:
        """
        unsigned_tx_c = self.get_data_file("tsx_uns_enc01.txt")
        creds = self.get_creds()
        unsigned_tx = await wallet.load_unsigned_tx(
            creds.view_key_private, unsigned_tx_c
        )
        await self.tx_sign_unsigned_msg(unsigned_tx, "tsx_uns_enc01.txt")

    async def test_tx_sign_uns01_boost_full_2dest_sub(self):
        """
        Testing tx signature, one input. full RCT, boost serialized
        :return:
        """
        unsigned_tx_c = self.get_data_file("tsx_uns_enc02.txt")
        creds = self.get_creds()
        unsigned_tx = await wallet.load_unsigned_tx(
            creds.view_key_private, unsigned_tx_c
        )
        await self.tx_sign_unsigned_msg(unsigned_tx, "tsx_uns_enc02.txt")

    async def test_trezor_ki(self):
        creds = self.get_trezor_creds(0)
        ki_data = self.get_data_file("ki_sync_01.txt")
        ki_loaded = await wallet.load_exported_outputs(
            creds.view_key_private, ki_data
        )

        self.assertEqual(ki_loaded.m_spend_public_key, crypto.encodepoint(creds.spend_key_public))
        self.assertEqual(ki_loaded.m_view_public_key, crypto.encodepoint(creds.view_key_public))

        tagent = self.init_agent(creds=creds)
        res = await tagent.import_outputs(ki_loaded.tds)
        await self.verify_ki_export(res, ki_loaded)

    async def test_trezor_txs(self):
        if os.getenv('SKIP_TREZOR_TSX', False):
            self.skipTest('Skipped by ENV var')

        files = self.get_trezor_tsx_tests()
        creds = self.get_trezor_creds(0)
        all_creds = [self.get_trezor_creds(0), self.get_trezor_creds(1), self.get_trezor_creds(2)]
        
        for fl in files:
            with self.subTest(msg=fl):
                unsigned_tx_c = self.get_data_file(fl)
                unsigned_tx = await wallet.load_unsigned_tx(
                    creds.view_key_private, unsigned_tx_c
                )

                tagent = self.init_agent(creds=creds)
                await self.tx_sign_test(tagent, unsigned_tx, creds, all_creds, fl)

    async def tx_sign_unsigned_msg(self, unsigned_tx, fl=None):
        """
        Signs tx stored in unsigned tx message
        :param unsigned_tx:
        :param fl:
        :return:
        """
        tagent = self.init_agent()
        await self.tx_sign_test(tagent, unsigned_tx, self.get_creds(),
                                [self.get_creds(), self.get_creds_01(), self.get_creds_02()], fl)

    async def tx_sign_unsigned(self, unsigned_tx, fl=None):
        """
        Tx sign test with given unsigned transaction data
        :param unsigned_tx:
        :param fl:
        :return:
        """
        reader = xmrserialize.MemoryReaderWriter(bytearray(unsigned_tx))
        ar = xmrserialize.Archive(reader, False)
        unsig = xmrtypes.UnsignedTxSet()
        await ar.message(unsig)
        await self.tx_sign_unsigned_msg(unsig, fl)

    async def tx_sign_unsigned_boost(self, unsigned_tx, fl=None):
        """
        Tx sign test with given unsigned transaction data, serialized by boost -
        unsigned tx produced by watch-only cli wallet.

        :param unsigned_tx:
        :param fl:
        :return:
        """
        reader = xmrserialize.MemoryReaderWriter(bytearray(unsigned_tx))
        ar = xmrboost.Archive(reader, False)
        unsig = xmrtypes.UnsignedTxSet()
        await ar.root()
        await ar.message(unsig)
        await self.tx_sign_unsigned_msg(unsig, fl)

    async def tx_sign_pending_boost(self, pending_tx, fl=None):
        """
        Signs transaction produced by the wallet-rpc, metadata parser, boost
        :param metadata:
        :param fl:
        :return:
        """
        reader = xmrserialize.MemoryReaderWriter(bytearray(pending_tx))
        ar = xmrboost.Archive(reader, False)
        pending = xmrtypes.PendingTransaction()
        await ar.root()
        await ar.message(pending)

        tagent = self.init_agent()
        await self.tx_sign_test(tagent, pending.construction_data, self.get_creds(),
                                [self.get_creds(), self.get_creds_01(), self.get_creds_02()], fl, sign_tx=True)

    def init_trezor(self, creds=None):
        """
        Initialize new trezor instance
        :type creds: object
        :return:
        """
        trez = token.TokenLite()
        trez.creds = self.get_creds() if creds is None else creds
        return trez

    def init_agent(self, creds=None):
        """
        Initialize new agent instance
        :return:
        """
        return agent_lite.Agent(self.init_trezor(creds=creds))


if __name__ == "__main__":
    unittest.main()  # pragma: no cover



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class DebugLinkMemory(p.MessageType):
    MESSAGE_WIRE_TYPE = 111
    FIELDS = {
        1: ('memory', p.BytesType, 0),
    }

    def __init__(
        self,
        memory: bytes = None,
    ) -> None:
        self.memory = memory



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class MoneroTransactionMlsagDoneAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 508
    FIELDS = {
        1: ('full_message_hash', p.BytesType, 0),
    }

    def __init__(
        self,
        full_message_hash: bytes = None,
    ) -> None:
        self.full_message_hash = full_message_hash



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class CardanoTxRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 304
    FIELDS = {
        1: ('tx_index', p.UVarintType, 0),
        2: ('tx_hash', p.BytesType, 0),
        3: ('tx_body', p.BytesType, 0),
    }

    def __init__(
        self,
        tx_index: int = None,
        tx_hash: bytes = None,
        tx_body: bytes = None,
    ) -> None:
        self.tx_index = tx_index
        self.tx_hash = tx_hash
        self.tx_body = tx_body



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
from .MoneroKeyImageExportInitRequest import MoneroKeyImageExportInitRequest
from .MoneroKeyImageSyncFinalRequest import MoneroKeyImageSyncFinalRequest
from .MoneroKeyImageSyncStepRequest import MoneroKeyImageSyncStepRequest


class MoneroKeyImageSyncRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 511
    FIELDS = {
        1: ('init', MoneroKeyImageExportInitRequest, 0),
        2: ('step', MoneroKeyImageSyncStepRequest, 0),
        3: ('final_msg', MoneroKeyImageSyncFinalRequest, 0),
    }

    def __init__(
        self,
        init: MoneroKeyImageExportInitRequest = None,
        step: MoneroKeyImageSyncStepRequest = None,
        final_msg: MoneroKeyImageSyncFinalRequest = None,
    ) -> None:
        self.init = init
        self.step = step
        self.final_msg = final_msg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import logging


logger = logging.getLogger(__name__)


def debug(name, msg, *args):
    logger = logging.getLogger(name)
    logger.debug(msg, *args)


def info(name, msg, *args):
    logger = logging.getLogger(name)
    logger.info(msg, *args)


def warning(name, msg, *args):
    logger = logging.getLogger(name)
    logger.warning(msg, *args)


def error(name, msg, *args):
    logger = logging.getLogger(name)
    logger.error(msg, *args)


def exception(name, exc):
    logger = logging.getLogger(name)
    logger.error(exc)


def critical(name, msg, *args):
    logger = logging.getLogger(name)
    logger.error(msg, *args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018


def collect(*args, **kwargs):
    pass


def mem_free(*args, **kwargs):
    return 1000


def mem_alloc(*args, **kwargs):
    return 100



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore
from .MoneroTransactionDestinationEntry import MoneroTransactionDestinationEntry


class MoneroTransactionData(p.MessageType):
    FIELDS = {
        1: ('version', p.UVarintType, 0),
        2: ('payment_id', p.BytesType, 0),
        3: ('unlock_time', p.UVarintType, 0),
        4: ('outputs', MoneroTransactionDestinationEntry, p.FLAG_REPEATED),
        5: ('change_dts', MoneroTransactionDestinationEntry, 0),
        6: ('num_inputs', p.UVarintType, 0),
        7: ('mixin', p.UVarintType, 0),
        8: ('fee', p.UVarintType, 0),
        9: ('account', p.UVarintType, 0),
        10: ('minor_indices', p.UVarintType, p.FLAG_REPEATED),
        11: ('is_multisig', p.BoolType, 0),
        12: ('exp_tx_prefix_hash', p.BytesType, 0),
        13: ('use_tx_keys', p.BytesType, p.FLAG_REPEATED),
        14: ('is_bulletproof', p.BoolType, 0),
    }

    def __init__(
        self,
        version: int = None,
        payment_id: bytes = None,
        unlock_time: int = None,
        outputs: List[MoneroTransactionDestinationEntry] = None,
        change_dts: MoneroTransactionDestinationEntry = None,
        num_inputs: int = None,
        mixin: int = None,
        fee: int = None,
        account: int = None,
        minor_indices: List[int] = None,
        is_multisig: bool = None,
        exp_tx_prefix_hash: bytes = None,
        use_tx_keys: List[bytes] = None,
        is_bulletproof: bool = None,
    ) -> None:
        self.version = version
        self.payment_id = payment_id
        self.unlock_time = unlock_time
        self.outputs = outputs if outputs is not None else []
        self.change_dts = change_dts
        self.num_inputs = num_inputs
        self.mixin = mixin
        self.fee = fee
        self.account = account
        self.minor_indices = minor_indices if minor_indices is not None else []
        self.is_multisig = is_multisig
        self.exp_tx_prefix_hash = exp_tx_prefix_hash
        self.use_tx_keys = use_tx_keys if use_tx_keys is not None else []
        self.is_bulletproof = is_bulletproof



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore
from .MoneroSubAddressIndicesList import MoneroSubAddressIndicesList


class MoneroKeyImageExportInitRequest(p.MessageType):
    FIELDS = {
        1: ('num', p.UVarintType, 0),
        2: ('hash', p.BytesType, 0),
        3: ('address_n', p.UVarintType, p.FLAG_REPEATED),
        4: ('network_type', p.UVarintType, 0),
        5: ('subs', MoneroSubAddressIndicesList, p.FLAG_REPEATED),
    }

    def __init__(
        self,
        num: int = None,
        hash: bytes = None,
        address_n: List[int] = None,
        network_type: int = None,
        subs: List[MoneroSubAddressIndicesList] = None,
    ) -> None:
        self.num = num
        self.hash = hash
        self.address_n = address_n if address_n is not None else []
        self.network_type = network_type
        self.subs = subs if subs is not None else []



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class MoneroKeyImageSyncFinalAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 522
    FIELDS = {
        1: ('enc_key', p.BytesType, 0),
    }

    def __init__(
        self,
        enc_key: bytes = None,
    ) -> None:
        self.enc_key = enc_key



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class RippleSignedTx(p.MessageType):
    MESSAGE_WIRE_TYPE = 403
    FIELDS = {
        1: ('signature', p.BytesType, 0),
        2: ('serialized_tx', p.BytesType, 0),
    }

    def __init__(
        self,
        signature: bytes = None,
        serialized_tx: bytes = None,
    ) -> None:
        self.signature = signature
        self.serialized_tx = serialized_tx



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import unittest

import aiounittest
from monero_glue.xmr import common, crypto
from monero_glue.xmr.core import ec_py


class CryptoTest(aiounittest.AsyncTestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(CryptoTest, self).__init__(*args, **kwargs)

    def init_sc(self, x):
        if x < (1 << 64):
            return crypto.sc_init(x)
        return crypto.decodeint(ec_py.encodeint(x))

    def test_ed_crypto(self):
        sqr = ec_py.fe_expmod(ec_py.py_fe_sqrtm1, 2)
        self.assertEqual(sqr, ec_py.fe_mod(-1))
        self.assertEqual(
            ec_py.py_fe_A, ec_py.fe_mod(2 * (1 - ec_py.d) * ec_py.inv(1 + ec_py.py_d))
        )

        self.assertEqual(
            ec_py.fe_expmod(ec_py.py_fe_fffb1, 2),
            ec_py.fe_mod(-2 * ec_py.py_fe_A * (ec_py.py_fe_A + 2)),
        )
        self.assertEqual(
            ec_py.fe_expmod(ec_py.py_fe_fffb2, 2),
            ec_py.fe_mod(2 * ec_py.py_fe_A * (ec_py.py_fe_A + 2)),
        )
        self.assertEqual(
            ec_py.fe_expmod(ec_py.py_fe_fffb3, 2),
            ec_py.fe_mod(-ec_py.py_fe_sqrtm1 * ec_py.py_fe_A * (ec_py.py_fe_A + 2)),
        )
        self.assertEqual(
            ec_py.fe_expmod(ec_py.py_fe_fffb4, 2),
            ec_py.fe_mod(ec_py.py_fe_sqrtm1 * ec_py.py_fe_A * (ec_py.py_fe_A + 2)),
        )

    def test_encoding(self):
        point = bytes(
            [
                0x24,
                0x86,
                0x22,
                0x47,
                0x97,
                0xd0,
                0x5c,
                0xae,
                0x3c,
                0xba,
                0x4b,
                0xe0,
                0x43,
                0xbe,
                0x2d,
                0xb0,
                0xdf,
                0x38,
                0x1f,
                0x3f,
                0x19,
                0xcf,
                0xa1,
                0x13,
                0xf8,
                0x6a,
                0xb3,
                0x8e,
                0x3d,
                0x8d,
                0x2b,
                0xd0,
            ]
        )
        self.assertEqual(point, crypto.encodepoint(crypto.decodepoint(point)))
        self.assertTrue(
            crypto.point_eq(
                crypto.decodepoint(point),
                crypto.decodepoint(crypto.encodepoint(crypto.decodepoint(point))),
            )
        )

    def test_scalarmult_base(self):
        scalar = crypto.decodeint(
            bytes(
                [
                    0xa0,
                    0xee,
                    0xa4,
                    0x91,
                    0x40,
                    0xa3,
                    0xb0,
                    0x36,
                    0xda,
                    0x30,
                    0xea,
                    0xcf,
                    0x64,
                    0xbd,
                    0x9d,
                    0x56,
                    0xce,
                    0x3e,
                    0xf6,
                    0x8b,
                    0xa8,
                    0x2e,
                    0xf1,
                    0x35,
                    0x71,
                    0xec,
                    0x51,
                    0x1e,
                    0xdb,
                    0xcf,
                    0x83,
                    0x03,
                ]
            )
        )
        exp = bytes(
            [
                0x16,
                0xbb,
                0x4a,
                0x3c,
                0x44,
                0xe2,
                0xce,
                0xd5,
                0x11,
                0xfc,
                0x0d,
                0x4c,
                0xd8,
                0x6b,
                0x13,
                0xb3,
                0xaf,
                0x21,
                0xef,
                0xc9,
                0x9f,
                0xb0,
                0x35,
                0x61,
                0x99,
                0xfa,
                0xc4,
                0x89,
                0xf2,
                0x54,
                0x4c,
                0x09,
            ]
        )
        res = crypto.scalarmult_base(scalar)
        self.assertEqual(exp, crypto.encodepoint(res))
        self.assertTrue(crypto.point_eq(crypto.decodepoint(exp), res))

        scalar = crypto.decodeint(
            bytes(
                [
                    0xfd,
                    0x29,
                    0x0d,
                    0xce,
                    0x39,
                    0xf7,
                    0x81,
                    0xae,
                    0xbb,
                    0xdb,
                    0xd2,
                    0x45,
                    0x84,
                    0xed,
                    0x6d,
                    0x48,
                    0xbd,
                    0x30,
                    0x0d,
                    0xe1,
                    0x9d,
                    0x9c,
                    0x3d,
                    0xec,
                    0xfd,
                    0xa0,
                    0xa6,
                    0xe2,
                    0xc6,
                    0x75,
                    0x1d,
                    0x0f,
                ]
            )
        )
        exp = bytes(
            [
                0x12,
                0x3d,
                0xaf,
                0x90,
                0xfc,
                0x26,
                0xf1,
                0x3c,
                0x65,
                0x29,
                0xe6,
                0xb4,
                0x9b,
                0xfe,
                0xd4,
                0x98,
                0x99,
                0x5a,
                0xc3,
                0x83,
                0xef,
                0x19,
                0xc0,
                0xdb,
                0x67,
                0x71,
                0x14,
                0x3f,
                0x24,
                0xba,
                0x8d,
                0xd5,
            ]
        )
        res = crypto.scalarmult_base(scalar)
        self.assertEqual(exp, crypto.encodepoint(res))
        self.assertTrue(crypto.point_eq(crypto.decodepoint(exp), res))

    def test_scalarmult(self):
        priv = bytes(
            [
                0x34,
                0x82,
                0xfb,
                0x97,
                0x35,
                0xef,
                0x87,
                0x9f,
                0xca,
                0xe5,
                0xec,
                0x77,
                0x21,
                0xb5,
                0xd3,
                0x64,
                0x6e,
                0x15,
                0x5c,
                0x4f,
                0xb5,
                0x8d,
                0x6c,
                0xc1,
                0x1c,
                0x73,
                0x2c,
                0x9c,
                0x9b,
                0x76,
                0x62,
                0x0a,
            ]
        )
        pub = bytes(
            [
                0x24,
                0x86,
                0x22,
                0x47,
                0x97,
                0xd0,
                0x5c,
                0xae,
                0x3c,
                0xba,
                0x4b,
                0xe0,
                0x43,
                0xbe,
                0x2d,
                0xb0,
                0xdf,
                0x38,
                0x1f,
                0x3f,
                0x19,
                0xcf,
                0xa1,
                0x13,
                0xf8,
                0x6a,
                0xb3,
                0x8e,
                0x3d,
                0x8d,
                0x2b,
                0xd0,
            ]
        )
        exp = bytes(
            [
                0xad,
                0xcd,
                0x1f,
                0x58,
                0x81,
                0xf4,
                0x6f,
                0x25,
                0x49,
                0x00,
                0xa0,
                0x3c,
                0x65,
                0x4e,
                0x71,
                0x95,
                0x0a,
                0x88,
                0xa0,
                0x23,
                0x6f,
                0xa0,
                0xa3,
                0xa9,
                0x46,
                0xc9,
                0xb8,
                0xda,
                0xed,
                0x6e,
                0xf4,
                0x3d,
            ]
        )
        res = crypto.scalarmult(crypto.decodepoint(pub), crypto.decodeint(priv))
        self.assertEqual(exp, crypto.encodepoint(res))
        self.assertTrue(crypto.point_eq(crypto.decodepoint(exp), res))

    def test_cn_fast_hash(self):
        inp = bytes(
            [
                0x25,
                0x9e,
                0xf2,
                0xab,
                0xa8,
                0xfe,
                0xb4,
                0x73,
                0xcf,
                0x39,
                0x05,
                0x8a,
                0x0f,
                0xe3,
                0x0b,
                0x9f,
                0xf6,
                0xd2,
                0x45,
                0xb4,
                0x2b,
                0x68,
                0x26,
                0x68,
                0x7e,
                0xbd,
                0x6b,
                0x63,
                0x12,
                0x8a,
                0xff,
                0x64,
                0x05,
            ]
        )
        res = crypto.cn_fast_hash(inp)
        self.assertEqual(
            res,
            bytes(
                [
                    0x86,
                    0xdb,
                    0x87,
                    0xb8,
                    0x3f,
                    0xb1,
                    0x24,
                    0x6e,
                    0xfc,
                    0xa5,
                    0xf3,
                    0xb0,
                    0xdb,
                    0x09,
                    0xce,
                    0x3f,
                    0xa4,
                    0xd6,
                    0x05,
                    0xb0,
                    0xd1,
                    0x0e,
                    0x65,
                    0x07,
                    0xca,
                    0xc2,
                    0x53,
                    0xdd,
                    0x31,
                    0xa3,
                    0xec,
                    0x16,
                ]
            ),
        )

    def test_hash_to_scalar(self):
        inp = bytes(
            [
                0x25,
                0x9e,
                0xf2,
                0xab,
                0xa8,
                0xfe,
                0xb4,
                0x73,
                0xcf,
                0x39,
                0x05,
                0x8a,
                0x0f,
                0xe3,
                0x0b,
                0x9f,
                0xf6,
                0xd2,
                0x45,
                0xb4,
                0x2b,
                0x68,
                0x26,
                0x68,
                0x7e,
                0xbd,
                0x6b,
                0x63,
                0x12,
                0x8a,
                0xff,
                0x64,
                0x05,
            ]
        )
        res = crypto.hash_to_scalar(inp)
        exp = self.init_sc(
            0x6eca331dd53c2ca07650ed1b005d6a42aef0ffd0dfc092616124e255b920799
        )
        self.assertTrue(crypto.sc_eq(res, exp))

    def test_hash_to_point(self):
        data = bytes(
            [
                0x42,
                0xf6,
                0x83,
                0x5b,
                0xf8,
                0x31,
                0x14,
                0xa1,
                0xf5,
                0xf6,
                0x07,
                0x6f,
                0xe7,
                0x9b,
                0xdf,
                0xa0,
                0xbd,
                0x67,
                0xc7,
                0x4b,
                0x88,
                0xf1,
                0x27,
                0xd5,
                0x45,
                0x72,
                0xd3,
                0x91,
                0x0d,
                0xd0,
                0x92,
                0x01,
            ]
        )
        res = crypto.hash_to_ec(data)
        res_p = crypto.encodepoint(res)
        self.assertEqual(
            res_p,
            bytes(
                [
                    0x54,
                    0x86,
                    0x3a,
                    0x04,
                    0x64,
                    0xc0,
                    0x08,
                    0xac,
                    0xc9,
                    0x9c,
                    0xff,
                    0xb1,
                    0x79,
                    0xbc,
                    0x6c,
                    0xf3,
                    0x4e,
                    0xb1,
                    0xbb,
                    0xdf,
                    0x6c,
                    0x29,
                    0xf7,
                    0xa0,
                    0x70,
                    0xa7,
                    0xc6,
                    0x37,
                    0x6a,
                    0xe3,
                    0x0a,
                    0xb5,
                ]
            ),
        )

    def test_derivation_to_scalar(self):
        derivation = bytes(
            [
                0xe7,
                0x20,
                0xa0,
                0x9f,
                0x2e,
                0x3a,
                0x0b,
                0xbf,
                0x4e,
                0x4b,
                0xa7,
                0xad,
                0x93,
                0x65,
                0x3b,
                0xb2,
                0x96,
                0x88,
                0x55,
                0x10,
                0x12,
                0x1f,
                0x80,
                0x6a,
                0xcb,
                0x2a,
                0x5f,
                0x91,
                0x68,
                0xfa,
                0xfa,
                0x01,
            ]
        )
        scalar = bytes(
            [
                0x25,
                0xd0,
                0x87,
                0x63,
                0x41,
                0x4c,
                0x37,
                0x9a,
                0xa9,
                0xcf,
                0x98,
                0x9c,
                0xdc,
                0xb3,
                0xca,
                0xdd,
                0x36,
                0xbd,
                0x51,
                0x93,
                0xb5,
                0x00,
                0x10,
                0x7d,
                0x6b,
                0xf5,
                0xf9,
                0x21,
                0xf1,
                0x8e,
                0x47,
                0x0e,
            ]
        )
        sc_int = crypto.derivation_to_scalar(crypto.decodepoint(derivation), 0)
        self.assertEqual(scalar, crypto.encodeint(sc_int))

    def test_generate_key_derivation(self):
        key_pub = crypto.decodepoint(
            bytes(
                [
                    0x77,
                    0x39,
                    0xc9,
                    0x5d,
                    0x32,
                    0x98,
                    0xe2,
                    0xf8,
                    0x73,
                    0x62,
                    0xdb,
                    0xa9,
                    0xe0,
                    0xe0,
                    0xb3,
                    0x98,
                    0x0a,
                    0x69,
                    0x2a,
                    0xe8,
                    0xe2,
                    0xf1,
                    0x67,
                    0x96,
                    0xb0,
                    0xe3,
                    0x82,
                    0x09,
                    0x8c,
                    0xd6,
                    0xbd,
                    0x83,
                ]
            )
        )
        key_priv = crypto.decodeint(
            bytes(
                [
                    0x34,
                    0x82,
                    0xfb,
                    0x97,
                    0x35,
                    0xef,
                    0x87,
                    0x9f,
                    0xca,
                    0xe5,
                    0xec,
                    0x77,
                    0x21,
                    0xb5,
                    0xd3,
                    0x64,
                    0x6e,
                    0x15,
                    0x5c,
                    0x4f,
                    0xb5,
                    0x8d,
                    0x6c,
                    0xc1,
                    0x1c,
                    0x73,
                    0x2c,
                    0x9c,
                    0x9b,
                    0x76,
                    0x62,
                    0x0a,
                ]
            )
        )
        deriv_exp = bytes(
            [
                0xfa,
                0x18,
                0x8a,
                0x45,
                0xa0,
                0xe4,
                0xda,
                0xcc,
                0xc0,
                0xe6,
                0xd4,
                0xf6,
                0xf6,
                0x85,
                0x8f,
                0xd4,
                0x63,
                0x92,
                0x10,
                0x4b,
                0xe7,
                0x41,
                0x83,
                0xec,
                0x00,
                0x47,
                0xe7,
                0xe9,
                0xf4,
                0xea,
                0xf7,
                0x39,
            ]
        )
        self.assertEqual(
            deriv_exp,
            crypto.encodepoint(crypto.generate_key_derivation(key_pub, key_priv)),
        )

    def test_h(self):
        H = bytes(
            [
                0x8b,
                0x65,
                0x59,
                0x70,
                0x15,
                0x37,
                0x99,
                0xaf,
                0x2a,
                0xea,
                0xdc,
                0x9f,
                0xf1,
                0xad,
                0xd0,
                0xea,
                0x6c,
                0x72,
                0x51,
                0xd5,
                0x41,
                0x54,
                0xcf,
                0xa9,
                0x2c,
                0x17,
                0x3a,
                0x0d,
                0xd3,
                0x9c,
                0x1f,
                0x94,
            ]
        )
        self.assertEqual(crypto.encodepoint(crypto.gen_H()), H)

    def test_h_pow(self):
        hp = crypto.gen_Hpow(10)
        self.assertEqual(crypto.encodepoint(hp[0]), crypto.encodepoint(crypto.gen_H()))
        for i in range(1, 10):
            crypto.check_ed25519point(hp[i])
            self.assertEqual(
                crypto.encodepoint(hp[i]),
                crypto.encodepoint(
                    crypto.scalarmult(crypto.gen_H(), crypto.sc_init(2 ** i))
                ),
            )

    def test_signature(self):
        for i in range(10):
            priv = crypto.random_scalar()
            data = crypto.cn_fast_hash(bytes(bytearray([i])))

            c, r, pub = crypto.generate_signature(data, priv)
            res = crypto.check_signature(data, c, r, pub)
            self.assertEqual(res, 1)

            res2 = crypto.check_signature(
                data, crypto.sc_add(c, crypto.sc_init(1)), r, pub
            )
            self.assertEqual(res2, 0)

    def test_edhex(self):
        inputs = [crypto.q - 2 ** 9, crypto.q - 10, 0, 100, 2 ** 200 + 10] + [
            common.rand.randrange(0, crypto.q - 2) for _ in range(20)
        ]

        for x in inputs:
            l = crypto.encode_ed25519(x)
            d = crypto.decode_ed25519(l)
            self.assertEqual(x, d)

    def test_modm(self):
        inputs = [crypto.l - 2 ** 9, crypto.l - 10, 0, 100, 2 ** 200 + 10] + [
            common.rand.randrange(0, crypto.l - 2) for _ in range(20)
        ]

        for x in inputs:
            l = crypto.encode_modm(x)
            d = crypto.decode_modm(l)
            self.assertEqual(x, d)

    def test_ge25519_double_scalarmult_vartime2(self):
        for i in range(10):
            ap = crypto.random_scalar()
            bp = crypto.random_scalar()
            A = crypto.scalarmult_base(ap)
            B = crypto.scalarmult_base(bp)
            a = crypto.random_scalar()
            b = crypto.random_scalar()

            R = crypto.ge_double_scalarmult_base_vartime2(a, A, b, B)
            R_exp = crypto.point_add(crypto.scalarmult(A, a), crypto.scalarmult(B, b))
            self.assertTrue(crypto.point_eq(R, R_exp))

    def test_ge25519_double_scalarmult_vartime(self):
        for i in range(10):
            ap = crypto.random_scalar()
            A = crypto.scalarmult_base(ap)
            a = crypto.random_scalar()
            b = crypto.random_scalar()

            R = crypto.ge_double_scalarmult_base_vartime(a, A, b)
            R_exp = crypto.point_add(crypto.scalarmult(A, a), crypto.scalarmult_base(b))
            self.assertTrue(crypto.point_eq(R, R_exp))


if __name__ == "__main__":
    unittest.main()  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import logging
import traceback

from monero_glue.agent import agent_misc
from monero_glue.hwtoken import misc as tmisc
from monero_glue.messages import (
    MoneroGetAddress,
    MoneroGetWatchKey,
    MoneroKeyImageSyncRequest,
    MoneroKeyImageSyncFinalRequest,
    MoneroKeyImageSyncStepRequest,
    MoneroTransactionAllOutSetRequest,
    MoneroTransactionAllOutSetAck,
    MoneroTransactionFinalRequest,
    MoneroTransactionFinalAck,
    MoneroTransactionInitRequest,
    MoneroTransactionInitAck,
    MoneroTransactionInputsPermutationRequest,
    MoneroTransactionInputViniRequest,
    MoneroTransactionMlsagDoneRequest,
    MoneroTransactionMlsagDoneAck,
    MoneroTransactionSetInputRequest,
    MoneroTransactionSetInputAck,
    MoneroTransactionSetOutputRequest,
    MoneroTransactionSetOutputAck,
    MoneroTransactionSignRequest,
    MoneroTransactionSignInputRequest,
    MoneroTransactionSignInputAck,
    Failure,
)
from monero_glue.protocol_base.base import TError
from monero_glue.xmr import common, key_image, monero, ring_ct
from monero_glue.xmr.tsx_data import TsxData
from monero_glue.xmr.enc import chacha_poly
from monero_serialize import xmrserialize, xmrtypes

logger = logging.getLogger(__name__)


DEFAULT_MONERO_BIP44 = [
    0x8000002c,
    0x80000080,
    0x80000000,
    0,
    0,
]  # parse_path(monero.DEFAULT_BIP32_PATH)


class TData(object):
    """
    Agent transaction-scoped data
    """

    def __init__(self):
        self.tsx_data = None  # type: TsxData
        self.tx_data = None  # construction data
        self.tx = xmrtypes.Transaction(version=2, vin=[], vout=[], extra=[])
        self.tx_in_hmacs = []
        self.tx_out_entr_hmacs = []
        self.tx_out_hmacs = []
        self.tx_out_rsigs = []
        self.tx_out_pk = []
        self.tx_out_ecdh = []
        self.source_permutation = []
        self.alphas = []
        self.spend_encs = []
        self.pseudo_outs = []
        self.couts = []
        self.tx_prefix_hash = None
        self.enc_salt1 = None
        self.enc_salt2 = None
        self.enc_keys = None  # encrypted tx keys


class Agent(object):
    """
    Glue agent, running on host
    """

    def __init__(self, trezor, address_n=None, network_type=None, **kwargs):
        self.trezor = trezor
        self.ct = None  # type: TData
        self.address_n = address_n if address_n else DEFAULT_MONERO_BIP44
        self.network_type = network_type

    def is_simple(self, rv):
        """
        True if simpe
        :param rv:
        :return:
        """
        return rv.type in [xmrtypes.RctType.Simple, xmrtypes.RctType.SimpleBulletproof]

    def is_bulletproof(self, rv):
        """
        True if bulletproof
        :param rv:
        :return:
        """
        return rv.type in [
            xmrtypes.RctType.FullBulletproof,
            xmrtypes.RctType.SimpleBulletproof,
        ]

    def is_error(self, response):
        """
        True if trezor returned an error
        :param response:
        :return:
        """
        return isinstance(response, TError) or isinstance(response, Failure)

    def handle_error(self, response):
        """
        Raises an error if Trezor returned an error.
        :param response:
        :return:
        """
        if not self.is_error(response):
            return
        raise agent_misc.TrezorReturnedError(resp=response)

    async def sign_unsigned_tx(self, unsig):
        """
        Processes unsigned transaction set, returns serialized signed transactions.
        :param unsig:
        :return:
        """
        txes = []
        for tx in unsig.txes:
            await self.sign_transaction_data(tx)
            txes.append(await self.serialized_tx())

        return txes

    async def sign_tx(self, construction_data):
        """
        Transfers transaction, serializes response

        :param construction_data:
        :return:
        """
        if not isinstance(construction_data, list):
            construction_data = [construction_data]

        txes = []
        for tdata in construction_data:
            await self.sign_transaction_data(tdata)
            txes.append(await self.serialized_tx())

        return txes

    async def serialized_tx(self):
        """
        Returns the last signed transaction as blob
        :return:
        """
        return await self.serialize_tx(self.ct.tx)

    async def serialize_tx(self, tx):
        """
        Serializes transaction
        :param tx:
        :return:
        """
        writer = xmrserialize.MemoryReaderWriter()
        ar1 = xmrserialize.Archive(writer, True)
        await ar1.message(tx, msg_type=xmrtypes.Transaction)
        return bytes(writer.get_buffer())

    async def sign_transaction_data(
        self, tx, multisig=False, exp_tx_prefix_hash=None, use_tx_keys=None
    ):
        """
        Uses Trezor to sign the transaction
        :param tx:
        :type tx: xmrtypes.TxConstructionData
        :param multisig:
        :param exp_tx_prefix_hash:
        :param use_tx_keys:
        :return:
        """
        self.ct = TData()
        self.ct.tx_data = tx

        payment_id = []
        extras = await monero.parse_extra_fields(tx.extra)
        extra_nonce = monero.find_tx_extra_field_by_type(extras, xmrtypes.TxExtraNonce)
        if extra_nonce and monero.has_encrypted_payment_id(extra_nonce.nonce):
            payment_id = monero.get_encrypted_payment_id_from_tx_extra_nonce(
                extra_nonce.nonce
            )
        elif extra_nonce and monero.has_payment_id(extra_nonce.nonce):
            payment_id = monero.get_payment_id_from_tx_extra_nonce(extra_nonce.nonce)

        # Init transaction
        tsx_data = TsxData()
        tsx_data.version = 1
        tsx_data.payment_id = payment_id
        tsx_data.unlock_time = tx.unlock_time
        tsx_data.outputs = tx.splitted_dsts
        tsx_data.change_dts = tx.change_dts
        tsx_data.num_inputs = len(tx.sources)
        tsx_data.mixin = len(tx.sources[0].outputs)
        tsx_data.fee = sum([x.amount for x in tx.sources]) - sum(
            [x.amount for x in tx.splitted_dsts]
        )
        tsx_data.account = tx.subaddr_account
        tsx_data.minor_indices = tx.subaddr_indices
        tsx_data.is_multisig = multisig
        tsx_data.is_bulletproof = False
        tsx_data.exp_tx_prefix_hash = common.defval(exp_tx_prefix_hash, b"")
        tsx_data.use_tx_keys = common.defval(use_tx_keys, [])
        self.ct.tx.unlock_time = tx.unlock_time

        self.ct.tsx_data = tsx_data
        tsx_data_pb = await tmisc.translate_tsx_data_pb(tsx_data)
        init_msg = MoneroTransactionInitRequest(
            version=0,
            address_n=self.address_n,
            network_type=self.network_type,
            tsx_data=tsx_data_pb,
        )

        t_res = await self.trezor.tsx_sign(
            MoneroTransactionSignRequest(init=init_msg)
        )  # type: MoneroTransactionInitAck
        self.handle_error(t_res)

        in_memory = t_res.in_memory
        self.ct.tx_out_entr_hmacs = t_res.hmacs

        # Set transaction inputs
        for idx, src in enumerate(tx.sources):
            src_bin = await tmisc.dump_msg(src)
            msg = MoneroTransactionSetInputRequest(src_entr=src_bin)

            t_res = await self.trezor.tsx_sign(
                MoneroTransactionSignRequest(set_input=msg)
            )  # type: MoneroTransactionSetInputAck
            self.handle_error(t_res)

            vini = await tmisc.parse_msg(t_res.vini, xmrtypes.TxinToKey())
            self.ct.tx.vin.append(vini)
            self.ct.tx_in_hmacs.append(t_res.vini_hmac)
            self.ct.pseudo_outs.append((t_res.pseudo_out, t_res.pseudo_out_hmac))
            self.ct.alphas.append(t_res.alpha_enc)
            self.ct.spend_encs.append(t_res.spend_enc)

        # Sort key image
        self.ct.source_permutation = list(range(len(tx.sources)))
        self.ct.source_permutation.sort(
            key=lambda x: self.ct.tx.vin[x].k_image, reverse=True
        )

        def swapper(x, y):
            self.ct.tx.vin[x], self.ct.tx.vin[y] = self.ct.tx.vin[y], self.ct.tx.vin[x]
            self.ct.tx_in_hmacs[x], self.ct.tx_in_hmacs[y] = (
                self.ct.tx_in_hmacs[y],
                self.ct.tx_in_hmacs[x],
            )
            self.ct.pseudo_outs[x], self.ct.pseudo_outs[y] = (
                self.ct.pseudo_outs[y],
                self.ct.pseudo_outs[x],
            )
            self.ct.alphas[x], self.ct.alphas[y] = self.ct.alphas[y], self.ct.alphas[x]
            self.ct.spend_encs[x], self.ct.spend_encs[y] = (
                self.ct.spend_encs[y],
                self.ct.spend_encs[x],
            )
            tx.sources[x], tx.sources[y] = tx.sources[y], tx.sources[x]

        common.apply_permutation(self.ct.source_permutation, swapper)

        if not in_memory:
            msg = MoneroTransactionInputsPermutationRequest(
                perm=self.ct.source_permutation
            )
            t_res = await self.trezor.tsx_sign(
                MoneroTransactionSignRequest(input_permutation=msg)
            )
            self.handle_error(t_res)

        # Set vin_i back - tx prefix hashing
        # Done only if not in-memory.
        if not in_memory:
            for idx in range(len(self.ct.tx.vin)):
                msg = MoneroTransactionInputViniRequest(
                    src_entr=await tmisc.dump_msg(tx.sources[idx]),
                    vini=await tmisc.dump_msg(self.ct.tx.vin[idx]),
                    vini_hmac=self.ct.tx_in_hmacs[idx],
                    pseudo_out=self.ct.pseudo_outs[idx][0] if not in_memory else None,
                    pseudo_out_hmac=self.ct.pseudo_outs[idx][1]
                    if not in_memory
                    else None,
                )
                t_res = await self.trezor.tsx_sign(
                    MoneroTransactionSignRequest(input_vini=msg)
                )
                self.handle_error(t_res)

        # Set transaction outputs
        for idx, dst in enumerate(tx.splitted_dsts):
            msg = MoneroTransactionSetOutputRequest(
                dst_entr=await tmisc.dump_msg(dst),
                dst_entr_hmac=self.ct.tx_out_entr_hmacs[idx],
            )
            t_res = await self.trezor.tsx_sign(
                MoneroTransactionSignRequest(set_output=msg)
            )  # type: MoneroTransactionSetOutputAck
            self.handle_error(t_res)

            self.ct.tx.vout.append(
                await tmisc.parse_msg(t_res.tx_out, xmrtypes.TxOut())
            )
            self.ct.tx_out_hmacs.append(t_res.vouti_hmac)
            self.ct.tx_out_rsigs.append(
                await tmisc.parse_msg(t_res.rsig, xmrtypes.RangeSig())
            )
            self.ct.tx_out_pk.append(
                await tmisc.parse_msg(t_res.out_pk, xmrtypes.CtKey())
            )
            self.ct.tx_out_ecdh.append(
                await tmisc.parse_msg(t_res.ecdh_info, xmrtypes.EcdhTuple())
            )

            # Rsig verification
            try:
                rsig = self.ct.tx_out_rsigs[-1]
                if not ring_ct.ver_range(C=None, rsig=rsig):
                    logger.warning("Rsing not valid")

            except Exception as e:
                logger.error("Exception rsig: %s" % e)
                traceback.print_exc()

        t_res = await self.trezor.tsx_sign(
            MoneroTransactionSignRequest(
                all_out_set=MoneroTransactionAllOutSetRequest()
            )
        )  # type: MoneroTransactionAllOutSetAck
        self.handle_error(t_res)

        rv = xmrtypes.RctSig()
        rv.p = xmrtypes.RctSigPrunable()
        rv.txnFee = t_res.rv.txn_fee
        rv.message = t_res.rv.message
        rv.type = t_res.rv.rv_type
        self.ct.tx.extra = list(bytearray(t_res.extra))

        # Verify transaction prefix hash correctness, tx hash in one pass
        self.ct.tx_prefix_hash = await monero.get_transaction_prefix_hash(self.ct.tx)
        if not common.ct_equal(t_res.tx_prefix_hash, self.ct.tx_prefix_hash):
            raise ValueError("Transaction prefix has does not match")

        # RctSig
        if self.is_simple(rv):
            if self.is_bulletproof(rv):
                rv.p.pseudoOuts = [x[0] for x in self.ct.pseudo_outs]
            else:
                rv.pseudoOuts = [x[0] for x in self.ct.pseudo_outs]

        # Range proof
        rv.p.rangeSigs = []
        rv.outPk = []
        rv.ecdhInfo = []
        for idx in range(len(self.ct.tx_out_rsigs)):
            rv.p.rangeSigs.append(self.ct.tx_out_rsigs[idx])
            rv.outPk.append(self.ct.tx_out_pk[idx])
            rv.ecdhInfo.append(self.ct.tx_out_ecdh[idx])

        # MLSAG message check
        t_res = await self.trezor.tsx_sign(
            MoneroTransactionSignRequest(mlsag_done=MoneroTransactionMlsagDoneRequest())
        )  # type: MoneroTransactionMlsagDoneAck
        self.handle_error(t_res)

        mlsag_hash = t_res.full_message_hash
        mlsag_hash_computed = await monero.get_pre_mlsag_hash(rv)
        if not common.ct_equal(mlsag_hash, mlsag_hash_computed):
            raise ValueError("Pre MLSAG hash has does not match")

        # Sign each input
        couts = []
        rv.p.MGs = []
        for idx, src in enumerate(tx.sources):
            msg = MoneroTransactionSignInputRequest(
                await tmisc.dump_msg(src),
                await tmisc.dump_msg(self.ct.tx.vin[idx]),
                self.ct.tx_in_hmacs[idx],
                self.ct.pseudo_outs[idx][0] if not in_memory else None,
                self.ct.pseudo_outs[idx][1] if not in_memory else None,
                self.ct.alphas[idx],
                self.ct.spend_encs[idx],
            )
            t_res = await self.trezor.tsx_sign(
                MoneroTransactionSignRequest(sign_input=msg)
            )  # type: MoneroTransactionSignInputAck
            self.handle_error(t_res)

            mg = await tmisc.parse_msg(t_res.signature, xmrtypes.MgSig())
            rv.p.MGs.append(mg)
            couts.append(t_res.cout)

        self.ct.tx.signatures = []
        self.ct.tx.rct_signatures = rv

        t_res = await self.trezor.tsx_sign(
            MoneroTransactionSignRequest(final_msg=MoneroTransactionFinalRequest())
        )  # type: MoneroTransactionFinalAck
        self.handle_error(t_res)

        if multisig:
            cout_key = t_res.cout_key
            for ccout in couts:
                self.ct.couts.append(
                    chacha_poly.decrypt(cout_key, ccout[0], ccout[1], ccout[2])
                )

        self.ct.enc_salt1, self.ct.enc_salt2 = t_res.salt, t_res.rand_mult
        self.ct.enc_keys = t_res.tx_enc_keys
        return self.ct.tx

    def last_transaction_data(self):
        """
        Returns last transaction data
        :return:
        """
        return self.ct

    async def get_address(self):
        """
        Get address from the device
        :return:
        """
        msg = MoneroGetAddress(address_n=self.address_n, network_type=self.network_type)
        res = await self.trezor.call(msg)
        return res

    async def get_watch_only(self):
        """
        Watch only key load
        :return:
        """
        msg = MoneroGetWatchKey(
            address_n=self.address_n, network_type=self.network_type
        )
        res = await self.trezor.get_view_key(msg)
        return res

    async def import_outputs(self, outputs):
        """
        Key images sync. Required for hot wallet be able to construct transactions.
        If the signed transaction is not relayed with the hot wallet it gets out of sync with
        key images. Thus importing is needed.

        Wallet2::import_outputs()

        :param outputs:
        :return:
        """
        ki_export_init = await key_image.generate_commitment(outputs)
        ki_export_init.address_n = self.address_n
        ki_export_init.network_type = self.network_type
        t_res = await self.trezor.key_image_sync(
            MoneroKeyImageSyncRequest(init=ki_export_init)
        )
        self.handle_error(t_res)

        sub_res = []
        iter = await key_image.yield_key_image_data(outputs)
        batches = common.chunk(iter, 10)
        for rr in batches:  # type: list[key_image.MoneroTransferDetails]
            t_res = await self.trezor.key_image_sync(
                MoneroKeyImageSyncRequest(step=MoneroKeyImageSyncStepRequest(tdis=rr))
            )
            self.handle_error(t_res)

            sub_res += t_res.kis

        t_res = await self.trezor.key_image_sync(
            MoneroKeyImageSyncRequest(final_msg=MoneroKeyImageSyncFinalRequest())
        )
        self.handle_error(t_res)

        # Decrypting phase
        enc_key = bytes(t_res.enc_key)
        final_res = []
        for sub in sub_res:  # type: key_image.MoneroExportedKeyImage
            plain = chacha_poly.decrypt(
                enc_key, bytes(sub.iv), bytes(sub.blob), bytes(sub.tag)
            )
            ki_bin = plain[:32]

            # ki = crypto.decodepoint(ki_bin)
            # sig = [crypto.decodeint(plain[32:64]), crypto.decodeint(plain[64:])]

            final_res.append((ki_bin, (plain[32:64], plain[64:])))

        return final_res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

#
# Note pickling is used for message serialization.
# This is just for the prototyping & fast PoC, pickling wont be used in the production.
# Instead, protobuf messages will be defined and parsed to avoid malicious pickling.
#

import argparse
import asyncio
import binascii
import collections
import json
import logging
import os
import pickle
import sys
import threading
import time
import traceback

from monero_glue import protobuf
from monero_glue.hwtoken import iface, token
from monero_glue.messages import MoneroWatchKey
from monero_glue.misc.bip import bip32
from monero_glue.protocol_base import messages
from monero_glue.xmr import crypto, monero, wallet
from monero_glue.xmr.core import mnemonic
from monero_poc.utils import cli
from monero_poc.utils import misc
from monero_serialize import xmrserialize

import coloredlogs
import eventlet
from eventlet import wsgi
from flask import Flask, abort, jsonify, request

logger = logging.getLogger(__name__)
coloredlogs.CHROOT_FILES = []
coloredlogs.install(level=logging.WARNING, use_chroot=False)
eventlet.monkey_patch(socket=True)


class TokenInterface(iface.TokenInterface):
    def __init__(self, server=None):
        self.server = server
        self.tsx_waiter = misc.CliPrompt(pre_wait_hook=server.update_prompt)
        self.ki_sync_waiter = misc.CliPrompt(pre_wait_hook=server.update_prompt)
        self.tsx_data = None
        self.confirmed_result = False

    def get_tx_data(self):
        return self.tsx_data

    @property
    def in_confirmation(self):
        return self.tsx_waiter.in_confirmation

    def confirmation(self, confirmed):
        self.confirmed_result = confirmed
        self.tsx_waiter.confirmation(confirmed)

    async def confirm_transaction(self, tsx_data, creds=None, ctx=None):
        self.confirmed_result = False
        self.tsx_data = tsx_data
        try:
            self.server.on_confirm_start()
            self.tsx_waiter.wait_confirmation()
            return self.confirmed_result

        finally:
            self.tsx_data = None

    async def transaction_signed(self, ctx=None):
        logger.debug("Transaction signed")

    async def transaction_error(self, *args, **kwargs):
        logger.error("Transaction error: %s %s" % (args, kwargs))

    async def transaction_finished(self, ctx=None):
        self.server.on_transaction_signed()

    async def transaction_step(self, step, sub_step=None, sub_step_total=None):
        logger.debug("Transaction step: %s, sub step: %s" % (step, sub_step))

    async def confirm_ki_sync(self, init_msg, ctx=None):
        self.server.on_confirm_ki_sync(init_msg)
        self.ki_sync_waiter.wait_confirmation()
        return self.ki_sync_waiter.confirmed_result

    async def ki_error(self, e, ctx=None):
        logger.error("ki sync error: %s" % e)

    async def ki_step(self, i, ctx=None):
        logger.debug("ki sync progress: %s" % i)

    async def ki_finished(self, ctx=None):
        logger.info("ki sync finished")


class TrezorServer(cli.BaseCli):
    """
    Trezor emulator server
    """

    prompt = "$> "

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.trez = None  # type: trezor_lite.TrezorLite
        self.args = None
        self.network_type = None
        self.creds = None  # type: monero.AccountCreds
        self.port = 46123
        self.trez_iface = TokenInterface(self)
        self.account_data = None
        self.ki_sync_data = None

        self.loop = asyncio.get_event_loop()
        self.running = True
        self.stop_event = threading.Event()
        self.local_data = threading.local()
        self.watch_only_waiter = misc.CliPrompt(pre_wait_hook=self.update_prompt)
        self.key_waiter = misc.CliPrompt(pre_wait_hook=self.update_prompt)
        self.ui_lock = threading.Lock()
        self.use_werkzeug = False
        self.thread_rest = None
        self.thread_loop = None
        self.rest_loop = None
        self.worker_loop = None

        self.debug = False
        self.server = None
        self.flask = Flask(__name__)

    #
    # CLI related
    #

    def update_intro(self):
        """
        Updates intro text for CLI header - adds version to it.
        :return:
        """
        self.intro = "-" * self.get_term_width() + "\n    Trezor server\n"

        if self.creds:
            self.intro += "\n    Account address: %s " % self.creds.address.decode(
                "utf8"
            )
        else:
            self.intro += "\n    Account not initialized, call new_wallet"

        self.intro += (
            ("\n    Running on: 127.0.0.1:%s " % self.port)
            + "\n"
            + "-" * self.get_term_width()
        )

    def update_prompt(self):
        """
        Prompt update
        :return:
        """
        flags = []
        if self.watch_only_waiter.in_confirmation:
            flags.append("W?")
        if self.key_waiter.in_confirmation:
            flags.append("Key?")
        if self.trez_iface.in_confirmation:
            flags.append("T?")
        if self.trez_iface.ki_sync_waiter.in_confirmation:
            flags.append("K?")

        flags_str = "|".join(flags)
        flags_suffix = "|" + flags_str if len(flags_str) > 0 else ""

        addr_str = self.creds.address.decode("ascii")[:6] if self.creds else "UNINIT"
        self.prompt = "[trezor %s%s]: " % (addr_str, flags_suffix)

    #
    # Server related
    #

    def shutdown_server(self):
        """
        Shutdown flask server
        :return:
        """
        func = request.environ.get("werkzeug.server.shutdown")
        if func is None:
            raise RuntimeError("Not running with the Werkzeug Server")
        func()

    def terminating(self):
        """
        Set state to terminating
        :return:
        """
        self.running = False
        self.stop_event.set()

    def wait_coro(self, coro):
        """
        Waits on the coroutine, for flask server
        :param coro:
        :return:
        """
        # Running on a different thread is not supported in flask:
        # future = asyncio.run_coroutine_threadsafe(self.on_ping(request=request), self.worker_loop)
        # return future.result()

        # Running on the common rest thread
        # return self.rest_loop.run_until_complete(self.on_ping(request=request))

        # Option: create a new loop as this worker is inside own thread
        loop = asyncio.new_event_loop()
        res = loop.run_until_complete(coro)
        loop.close()
        return res

    def init_rest(self):
        """
        Initializes rest server
        :return:
        """

        @self.flask.route("/api/v1.0/ping", methods=["GET"])
        def keep_alive():
            return self.wait_coro(self.on_ping(request=request))

        @self.flask.route("/api/v1.0/watch_only", methods=["GET", "POST"])
        def watch_only():
            return self.wait_coro(self.on_watch_only(request=request))

        @self.flask.route("/api/v1.0/get_keys", methods=["GET", "POST"])
        def get_keys():
            return self.wait_coro(self.on_keys(request=request))

        @self.flask.route("/api/v1.0/tx_sign", methods=["GET", "POST"])
        def tx_sign():
            return self.wait_coro(self.on_tx_sign(request=request))

        @self.flask.route("/api/v1.0/ki_sync", methods=["GET", "POST"])
        def ki_sync():
            return self.wait_coro(self.on_ki_sync(request=request))

    def wsgi_options(self):
        """
        Returns kwargs for wsgi server
        :return:
        """
        kwargs = dict()
        if False:
            kwargs["minimum_chunk_size"] = 1
        return kwargs

    def serve_werkzeug(self):
        """
        Developer local server, not for production use
        :return:
        """
        r = self.flask.run(debug=self.debug, port=self.port, threaded=True)
        logger.info("Started werkzeug server: %s" % r)

    def serve_eventlet(self):
        """
        Eventlet server, fast async, for production use
        Warning: Eventlet is tricky to work concurrently with classical threading locks

        :return:
        """
        listener = eventlet.listen(("0.0.0.0", self.port))
        logger.info(
            "Starting Eventlet listener %s for Flask %s" % (listener, self.flask)
        )
        wsgi.server(listener, self.flask, **self.wsgi_options())

    #
    # Handlers
    #

    async def on_ping(self, request=None):
        """
        Simple ping
        :param request:
        :return:
        """
        return jsonify({"result": True})

    async def on_watch_only(self, request=None):
        """
        Exports watch only credentials
        :param request:
        :return:
        """
        if not self.creds:
            logger.warning(
                "Agent asks for watch-only credentials, Trezor not initialized"
            )
            return abort(406)

        if self.watch_only_waiter.in_confirmation:
            logger.warning("Agent asks for watch-only credentials concurrently")
            return abort(406)

        # Prompt user to confirm.
        self.on_watchonly()
        confirmed = self.watch_only_waiter.wait_confirmation()

        if not confirmed:
            logger.warning("Watch only rejected")
            return abort(403)

        logger.info("Returning watch only credentials...")
        res = MoneroWatchKey(
            watch_key=crypto.encodeint(self.creds.view_key_private),
            address=self.creds.address,
        )

        return jsonify({"result": True, "payload": await self.proto_res(res)})

    async def on_tx_sign(self, request=None):
        """
        TX sign
        :param request:
        :return:
        """
        js = request.json

        try:
            js = request.json
            msg = await self.unproto_req(js)

            if not self.trez:
                logger.warning("Transaction signing request on unitialized Trezor")
                return abort(404)

            try:
                res = await self.trez.tsx_sign(msg)
                return jsonify({"result": True, "payload": await self.proto_res(res)})

            except Exception as e:
                return jsonify({"result": False, "exc": e})

        except Exception as e:
            traceback.print_exc()

    async def on_ki_sync(self, request=None):
        """
        Key image sync
        :param request:
        :return:
        """
        try:
            js = request.json
            msg = await self.unproto_req(js)

            if not self.trez:
                logger.warning("KeyImage sync request on unitialized Trezor")
                return abort(404)

            try:
                res = await self.trez.key_image_sync(msg)
                return jsonify({"result": True, "payload": await self.proto_res(res)})

            except Exception as e:
                return jsonify({"result": False, "exc": e})

        except Exception as e:
            traceback.print_exc()

    async def unproto_req(self, req):
        if "payload" not in req:
            return None
        msg, msg_type = req["payload"]["msg"], req["payload"]["msg_type"]
        return await self.unproto_msg(binascii.unhexlify(msg.encode("utf8")), msg_type)

    async def unproto_msg(self, msg, msg_type):
        reader = xmrserialize.MemoryReaderWriter(bytearray(msg))
        return await protobuf.load_message(
            reader, messages.get_message_from_type(msg_type)
        )

    async def proto_res(self, res):
        writer = xmrserialize.MemoryReaderWriter()
        await protobuf.dump_message(writer, res)
        return {
            "msg": binascii.hexlify(bytes(writer.get_buffer())).decode("ascii"),
            "msg_type": messages.get_message_type(res),
        }

    def unpickle_args(self, js):
        return pickle.loads(binascii.unhexlify(js["payload"].encode("utf8")))

    def pickle_res(self, res):
        return binascii.hexlify(pickle.dumps(res)).decode("utf8")

    #
    # Terminal
    #

    do_q = quit
    do_Q = quit

    def do_new_wallet(self, line):
        self.create_wallet(line)

    def do_address(self, line):
        self.poutput(self.creds.address.decode("ascii"))

    def do_keys(self, line):
        print(
            "Spend key priv:   0x%s"
            % binascii.hexlify(crypto.encodeint(self.creds.spend_key_private)).decode(
                "ascii"
            )
        )
        print(
            "View key priv:    0x%s"
            % binascii.hexlify(crypto.encodeint(self.creds.view_key_private)).decode(
                "ascii"
            )
        )
        print("")
        print(
            "Spend key pub:    0x%s"
            % binascii.hexlify(crypto.encodepoint(self.creds.spend_key_public)).decode(
                "ascii"
            )
        )
        print(
            "View key pub:     0x%s"
            % binascii.hexlify(crypto.encodepoint(self.creds.view_key_public)).decode(
                "ascii"
            )
        )

    def on_watchonly(self):
        self.poutput("-" * 80)
        self.poutput("Watch-only request received\nEnter W to confirm/reject\n")

    def on_confirm_start(self):
        self.poutput("-" * 80)
        self.poutput("Transaction confirmation procedure\nEnter T to start\n")

    def on_transaction_signed(self):
        self.poutput("-" * 80)
        self.poutput("Transaction was successfully signed\n")

    def on_confirm_ki_sync(self, msg):
        self.ki_sync_data = msg
        self.poutput("-" * 80)
        self.poutput("Key image sync procedure\nEnter K to start\n")

    def conv_disp_amount(self, amount):
        return wallet.conv_disp_amount(amount)

    def do_T(self, line):
        if not self.trez_iface.in_confirmation:
            self.poutput("No transaction in progress")
            return

        tsx_data = self.trez_iface.get_tx_data()
        self.poutput("Confirming transaction:")
        self.poutput("- " * 40)
        if tsx_data.payment_id:
            self.poutput(
                "  Payment ID: %s"
                % binascii.hexlify(tsx_data.payment_id).decode("utf8")
            )

        self.poutput("  Unlock time: %s" % tsx_data.unlock_time)
        self.poutput("  UTXOs: %s" % tsx_data.num_inputs)
        self.poutput("  Mixin: %s" % tsx_data.mixin)

        chg = tsx_data.change_dts
        addr_chg = (
            monero.public_addr_encode(chg.addr, chg.is_subaddress, self.network_type)
            if chg
            else None
        )

        for idx, out in enumerate(tsx_data.outputs):
            addr = monero.public_addr_encode(
                out.addr, out.is_subaddress, self.network_type
            )
            if addr != addr_chg:
                self.poutput(
                    "  Output %2d: %12.8f to %s, sub: %s"
                    % (
                        idx,
                        self.conv_disp_amount(out.amount),
                        addr.decode("utf8"),
                        out.is_subaddress,
                    )
                )
            else:
                self.poutput(
                    "  Change:    %12.8f to %s, sub: %s"
                    % (
                        self.conv_disp_amount(out.amount),
                        addr.decode("utf8"),
                        out.is_subaddress,
                    )
                )

        self.poutput("  Fee: %.8f" % self.conv_disp_amount(tsx_data.fee))
        self.poutput("  Account: %s" % tsx_data.account)

        self.poutput("- " * 40)
        result = self.select(
            [(0, "Confirm the transaction"), (1, "Reject")],
            "Do you confirm the transaction? ",
        )
        self.poutput("\n")
        self.trez_iface.confirmation(result == 0)
        self.update_prompt()

    def do_W(self, line):
        if not self.watch_only_waiter.in_confirmation:
            self.poutput("No prompt")
            return

        result = self.select(
            [(0, "Confirm"), (1, "Reject")],
            "Do you confirm sending watch-only credentials to the client? ",
        )

        self.poutput("\n")
        self.watch_only_waiter.confirmation(result == 0)
        self.update_prompt()

    def do_KK(self, line):
        if not self.key_waiter.in_confirmation:
            self.poutput("No prompt")
            return

        result = self.select(
            [(0, "Confirm"), (1, "Reject")],
            "Do you confirm exporting sec credentials to the client? ",
        )

        self.poutput("\n")
        self.key_waiter.confirmation(result == 0)
        self.update_prompt()

    def do_K(self, line):
        if not self.trez_iface.ki_sync_waiter.in_confirmation:
            self.poutput("No prompt")
            return

        self.poutput("Agent asks to perform key image sync.")
        self.poutput("Syncing %s outputs" % self.ki_sync_data.num)

        result = self.select(
            [(0, "Confirm"), (1, "Reject")], "Do you want to proceed? "
        )

        self.poutput("\n")
        self.trez_iface.ki_sync_waiter.confirmation(result == 0)
        self.update_prompt()

    do_t = do_T
    do_w = do_W
    do_k = do_K

    #
    # Work
    #

    def loop_work(self, loop):
        """
        Looping worker
        :return:
        """
        asyncio.set_event_loop(loop)
        loop.run_forever()

    def rest_thread_work(self):
        """
        Main work method for the server - accepting incoming connections.
        :return:
        """
        logger.info(
            "REST thread started %s %s %s dbg: %s"
            % (os.getpid(), os.getppid(), threading.current_thread(), self.debug)
        )
        try:
            self.worker_loop = asyncio.new_event_loop()
            self.rest_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.rest_loop)

            worker = threading.Thread(target=self.loop_work, args=(self.worker_loop,))
            worker.setDaemon(True)
            worker.start()

            self.init_rest()
            if self.debug or self.use_werkzeug:
                self.serve_werkzeug()
            else:
                self.serve_eventlet()

            logger.info("Terminating flask: %s" % self.flask)

        except Exception as e:
            logger.error("Exception: %s" % e)
            logger.error(traceback.format_exc())

        self.terminating()
        logger.info("Work loop terminated")

    def rest_server_boot(self):
        """
        Boost the server
        :return:
        """
        self.thread_rest = threading.Thread(target=self.rest_thread_work, args=())
        self.thread_rest.setDaemon(True)
        self.thread_rest.start()

    def create_wallet(self, line):
        """
        Creates a new account
        :return:
        """
        if self.args.account_file:
            if os.path.exists(self.args.account_file):
                logger.error("Wallet file exists, could not overwrite")
                return

        print("Generating new wallet...")
        seed = crypto.random_bytes(32)

        wl = bip32.Wallet.from_master_secret(seed)
        seed_bip32_words, seed_bip32_words_indices = wl.to_seed_words()
        seed_bip32_b58 = wl.serialize_b58()

        # Generate private keys based on the gen mechanism. Bip44 path + Monero backward compatible
        data = wl.get_child_for_path("m/44'/128'/0'/0/0")
        to_hash = data.chain_code + binascii.unhexlify(data.private_key.get_key())

        # to_hash is initial seed in the Monero sense, recoverable from this seed
        hashed = crypto.cn_fast_hash(to_hash)
        electrum_words = " ".join(mnemonic.mn_encode(hashed))

        keys = monero.generate_monero_keys(hashed)
        spend_sec, spend_pub, view_sec, view_pub = keys

        print("Seed:             0x%s" % binascii.hexlify(seed).decode("ascii"))
        print("Seed bip39 words: %s" % " ".join(seed_bip32_words))
        print("Seed bip32 b58:   %s" % seed_bip32_b58)

        print("Seed Monero:      0x%s" % binascii.hexlify(hashed).decode("ascii"))
        print("Seed Monero wrds: %s" % electrum_words)

        print("")
        print(
            "Spend key priv:   0x%s"
            % binascii.hexlify(crypto.encodeint(spend_sec)).decode("ascii")
        )
        print(
            "View key priv:    0x%s"
            % binascii.hexlify(crypto.encodeint(view_sec)).decode("ascii")
        )
        print("")
        print(
            "Spend key pub:    0x%s"
            % binascii.hexlify(crypto.encodepoint(spend_pub)).decode("ascii")
        )
        print(
            "View key pub:     0x%s"
            % binascii.hexlify(crypto.encodepoint(view_pub)).decode("ascii")
        )

        self.init_with_keys(spend_sec, view_sec)
        print("")
        print("Address:          %s" % self.creds.address.decode("ascii"))

        self.account_data = collections.OrderedDict()
        self.account_data["seed"] = binascii.hexlify(seed).decode("ascii")
        self.account_data["spend_key"] = binascii.hexlify(
            crypto.encodeint(spend_sec)
        ).decode("ascii")
        self.account_data["view_key"] = binascii.hexlify(
            crypto.encodeint(view_sec)
        ).decode("ascii")
        self.account_data["meta"] = collections.OrderedDict(
            [
                ("addr", self.creds.address.decode("ascii")),
                ("bip44_seed", binascii.hexlify(seed).decode("ascii")),
                ("bip32_39_words", " ".join(seed_bip32_words)),
                ("bip32_b58", seed_bip32_b58),
                ("monero_seed", binascii.hexlify(hashed).decode("ascii")),
                ("monero_words", electrum_words),
            ]
        )

        if self.args.account_file:
            with open(self.args.account_file, "w+") as fh:
                json.dump(self.account_data, fh, indent=2)
        print("Wallet generated")
        self.update_prompt()

    async def open_account(self):
        """
        Handles account open / management
        :return:
        """
        self.update_intro()
        self.update_prompt()

        priv_spend_key = None
        priv_view_key = None

        if not self.args.account_file and not self.args.spend_key:
            logger.debug("No account file nor spend key. Please generate new account")
            return

        acc_file_exists = False
        if self.args.account_file:
            acc_file_exists = os.path.exists(self.args.account_file)

            if acc_file_exists and self.args.spend_key:
                logger.error("Account file exists, spend key is ignored")
            if acc_file_exists and self.args.view_key:
                logger.error("Account file exists, view key is ignored")
            if acc_file_exists:
                with open(self.args.account_file) as fh:
                    self.account_data = json.load(fh)
                priv_spend_key = crypto.decodeint(
                    binascii.unhexlify(self.account_data["spend_key"].encode("ascii"))
                )
                priv_view_key = crypto.decodeint(
                    binascii.unhexlify(self.account_data["view_key"].encode("ascii"))
                )

        if not acc_file_exists and self.args.spend_key:
            priv_view = self.args.view_key.encode("utf8")
            priv_spend = self.args.spend_key.encode("utf8")
            priv_view_key = crypto.b16_to_scalar(priv_view)
            priv_spend_key = crypto.b16_to_scalar(priv_spend)

            self.account_data = {"spend_key": priv_spend_key, "view_key": priv_view_key}
            with open(self.args.account_file, "w") as fh:
                json.dump(
                    {
                        "spend_key": priv_spend.decode("ascii"),
                        "view_key": priv_view.decode("ascii"),
                    },
                    fh,
                    indent=2,
                )

        if priv_spend_key and priv_view_key:
            self.init_with_keys(priv_spend_key, priv_view_key)

    def init_with_keys(self, priv_spend_key, priv_view_key):
        """
        Initializes Trezor classes with the private keys
        :return:
        """
        self.creds = monero.AccountCreds.new_wallet(
            priv_view_key, priv_spend_key, self.network_type
        )
        self.update_intro()

        self.trez = token.TokenLite()
        self.trez.creds = self.creds
        self.trez.iface = self.trez_iface
        self.update_prompt()

    async def entry(self):
        """
        pass
        :return:
        """
        if self.args.debug:
            coloredlogs.install(level=logging.DEBUG, use_chroot=False)

        self.network_type = monero.NetworkTypes.MAINNET
        if self.args.testnet:
            self.network_type = monero.NetworkTypes.TESTNET
        elif self.args.stagenet:
            self.network_type = monero.NetworkTypes.STAGENET

        await self.open_account()

        logging.info("Starting rest server...")
        self.rest_server_boot()
        time.sleep(1)

        self.cmdloop()
        self.terminating()

    async def main(self):
        """
        Entry point
        :return:
        """
        parser = argparse.ArgumentParser(description="Trezor server")

        parser.add_argument(
            "--testnet",
            dest="testnet",
            default=False,
            action="store_const",
            const=True,
            help="Testnet",
        )

        parser.add_argument(
            "--stagenet",
            dest="stagenet",
            default=False,
            action="store_const",
            const=True,
            help="Stagenet",
        )

        parser.add_argument(
            "--debug",
            dest="debug",
            default=False,
            action="store_const",
            const=True,
            help="Debug",
        )

        parser.add_argument(
            "--account-file",
            dest="account_file",
            help="Trezor account file to use / open",
        )

        parser.add_argument(
            "--view-key", dest="view_key", help="Hex coded private view key"
        )

        parser.add_argument(
            "--spend-key", dest="spend_key", help="Hex coded private spend key"
        )

        args_src = sys.argv
        self.args, unknown = parser.parse_known_args(args=args_src[1:])

        sys.argv = [args_src[0]]
        await self.entry()
        sys.argv = args_src


async def amain():
    agent = TrezorServer()
    await agent.main()


def main():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(amain())
    # loop.run_forever()
    loop.close()


if __name__ == "__main__":
    main()



from monero_serialize.xmrtypes import EcdhTuple

from monero_glue.xmr import crypto


def copy_ecdh(ecdh):
    """
    Clones ECDH tuple
    :param ecdh:
    :return:
    """
    return EcdhTuple(mask=ecdh.mask, amount=ecdh.amount)


def recode_ecdh(ecdh, encode=True):
    """
    In-place ecdhtuple recoding
    :param ecdh:
    :param encode: if true encodes to byte representation, otherwise decodes from byte representation
    :return:
    """
    recode_int = crypto.encodeint if encode else crypto.decodeint
    ecdh.mask = recode_int(ecdh.mask)
    ecdh.amount = recode_int(ecdh.amount)
    return ecdh


def recode_msg(mgs, encode=True):
    """
    Recodes MGs signatures from raw forms to bytearrays so it works with serialization
    :param rv:
    :param encode: if true encodes to byte representation, otherwise decodes from byte representation
    :return:
    """
    recode_int = crypto.encodeint if encode else crypto.decodeint
    recode_point = crypto.encodepoint if encode else crypto.decodepoint

    for idx in range(len(mgs)):
        mgs[idx].cc = recode_int(mgs[idx].cc)
        if hasattr(mgs[idx], "II") and mgs[idx].II:
            for i in range(len(mgs[idx].II)):
                mgs[idx].II[i] = recode_point(mgs[idx].II[i])

        for i in range(len(mgs[idx].ss)):
            for j in range(len(mgs[idx].ss[i])):
                mgs[idx].ss[i][j] = recode_int(mgs[idx].ss[i][j])
    return mgs



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
from .TezosContractID import TezosContractID


class TezosOperationCommon(p.MessageType):
    FIELDS = {
        1: ('branch', p.BytesType, 0),
        2: ('tag', p.UVarintType, 0),
        3: ('source', TezosContractID, 0),
        4: ('fee', p.UVarintType, 0),
        5: ('counter', p.UVarintType, 0),
        6: ('gas_limit', p.UVarintType, 0),
        7: ('storage_limit', p.UVarintType, 0),
    }

    def __init__(
        self,
        branch: bytes = None,
        tag: int = None,
        source: TezosContractID = None,
        fee: int = None,
        counter: int = None,
        gas_limit: int = None,
        storage_limit: int = None,
    ) -> None:
        self.branch = branch
        self.tag = tag
        self.source = source
        self.fee = fee
        self.counter = counter
        self.gas_limit = gas_limit
        self.storage_limit = storage_limit



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
from .TezosContractID import TezosContractID


class TezosTransactionType(p.MessageType):
    FIELDS = {
        1: ('amount', p.UVarintType, 0),
        2: ('destination', TezosContractID, 0),
        3: ('parameters', p.BytesType, 0),
    }

    def __init__(
        self,
        amount: int = None,
        destination: TezosContractID = None,
        parameters: bytes = None,
    ) -> None:
        self.amount = amount
        self.destination = destination
        self.parameters = parameters



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
from .StellarAssetType import StellarAssetType


class StellarPaymentOp(p.MessageType):
    MESSAGE_WIRE_TYPE = 211
    FIELDS = {
        1: ('source_account', p.UnicodeType, 0),
        2: ('destination_account', p.UnicodeType, 0),
        3: ('asset', StellarAssetType, 0),
        4: ('amount', p.SVarintType, 0),
    }

    def __init__(
        self,
        source_account: str = None,
        destination_account: str = None,
        asset: StellarAssetType = None,
        amount: int = None,
    ) -> None:
        self.source_account = source_account
        self.destination_account = destination_account
        self.asset = asset
        self.amount = amount



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
from .TransactionType import TransactionType


class TxAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 22
    FIELDS = {
        1: ('tx', TransactionType, 0),
    }

    def __init__(
        self,
        tx: TransactionType = None,
    ) -> None:
        self.tx = tx



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018


import argparse
import asyncio
import binascii
import logging

import coloredlogs

from monero_glue.misc.bip import bip32
from monero_glue.misc.bip import bip39
from monero_glue.misc.bip import bip39_deriv
from monero_glue.xmr import crypto, monero
from monero_glue.xmr.core import mnemonic
from monero_glue.xmr.sub.addr import encode_addr
from monero_glue.xmr.sub.seed import SeedDerivation
from monero_glue.xmr.sub.xmr_net import NetworkTypes, net_version

logger = logging.getLogger(__name__)
coloredlogs.CHROOT_FILES = []
coloredlogs.install(level=logging.WARNING, use_chroot=False)


parser = argparse.ArgumentParser(description="Monero seed utility")
parser.add_argument("input", metavar="mnemonics", nargs="*", help="Input")

parser.add_argument(
    "--seed",
    dest="seed",
    default=False,
    action="store_const",
    const=True,
    help="Input is hexcoded seed / master secret, input to Hnode derivation",
)

parser.add_argument(
    "--wlist-seed",
    dest="wlist_seed",
    default=False,
    action="store_const",
    const=True,
    help="Mnemonics converted to seed using the wordlist indices",
)

parser.add_argument(
    "--electrum-mnemonics",
    dest="electrum_mnemonics",
    default=False,
    action="store_const",
    const=True,
    help="Monero electrum mneomonics encoding Monero master secret seed",
)

parser.add_argument(
    "--monero-master",
    dest="monero_master",
    default=False,
    action="store_const",
    const=True,
    help="Monero master secret seed",
)


parser.add_argument(
    "--mainnet",
    dest="mainnet",
    default=False,
    action="store_const",
    const=True,
    help="Mainnet",
)

parser.add_argument(
    "--testnet",
    dest="testnet",
    default=False,
    action="store_const",
    const=True,
    help="Testnet",
)

parser.add_argument(
    "--stagenet",
    dest="stagenet",
    default=False,
    action="store_const",
    const=True,
    help="Stagenet",
)

parser.add_argument(
    "--subs",
    dest="subs",
    default=False,
    action="store_const",
    const=True,
    help="Compute 5x5 sub addresses given network",
)

parser.add_argument(
    "--debug",
    dest="debug",
    default=False,
    action="store_const",
    const=True,
    help="Debug",
)

args = parser.parse_args()


async def amain(args):
    mnems = []
    for w in args.input:
        mnems += w.split(" ")
    mnems = [x.strip().lower() for x in mnems]

    bip44_derived = True

    if sum([args.electrum_mnemonics, args.wlist_seed, args.monero_master, args.seed]) > 1:
        raise ValueError('Conflicting input options')

    if args.electrum_mnemonics:
        bip44_derived = False
        sd = SeedDerivation.from_monero_mnemonics(mnems)

    elif args.monero_master:
        bip44_derived = False
        seed = binascii.unhexlify(" ".join(mnems))
        sd = SeedDerivation.from_monero_seed(seed)

    elif args.seed:
        seed = binascii.unhexlify(" ".join(mnems))
        sd = SeedDerivation.from_master_seed(seed)

    else:
        sd = SeedDerivation.from_mnemonics(mnems, args.wlist_seed)

    main_addr = sd.creds(network_type=NetworkTypes.MAINNET)
    test_addr = sd.creds(network_type=NetworkTypes.TESTNET)
    stage_addr = sd.creds(network_type=NetworkTypes.STAGENET)

    if bip44_derived:
        print("Seed bip39 words: %s" % " ".join(mnems))
        print("Seed bip32 b58:   %s\n" % binascii.hexlify(sd.master_seed).decode("ascii"))

    print("Seed Monero:      %s" % binascii.hexlify(sd.hashed).decode("ascii"))
    print("Seed Monero wrds: %s\n" % sd.electrum_words)

    print(
        "Private spend key: %s"
        % binascii.hexlify(crypto.encodeint(sd.spend_sec)).decode("ascii")
    )
    print(
        "Private view key:  %s\n"
        % binascii.hexlify(crypto.encodeint(sd.view_sec)).decode("ascii")
    )

    print(
        "Public spend key:  %s"
        % binascii.hexlify(crypto.encodepoint(sd.spend_pub)).decode("ascii")
    )
    print(
        "Public view key:   %s\n"
        % binascii.hexlify(crypto.encodepoint(sd.view_pub)).decode("ascii")
    )

    print("Mainnet Address:   %s" % main_addr.address.decode("ascii"))
    print("Testnet Address:   %s" % test_addr.address.decode("ascii"))
    print("Stagenet Address:  %s" % stage_addr.address.decode("ascii"))

    if args.subs:
        if args.mainnet or (not args.testnet and not args.stagenet):
            print('Mainnet Sub addresses: ')
            print_sub_addresses(sd, NetworkTypes.MAINNET)
        if args.testnet:
            print('Testnet Sub addresses: ')
            print_sub_addresses(sd, NetworkTypes.TESTNET)
        if args.stagenet:
            print('Stagenet Sub addresses: ')
            print_sub_addresses(sd, NetworkTypes.STAGENET)


def print_sub_addresses(sd, net_type, major_max=5, minor_max=5):
    for major, minor, addr in gen_sub_address(sd, net_type, major_max, minor_max):
        print(" %d, %d: %s" % (major, minor, addr.decode('ascii')))


def gen_sub_address(sd, net_type, major_max, minor_max):
    for major in range(major_max):
        for minor in range(minor_max):
            if major == 0 and minor == 0:
                continue

            D, C = monero.generate_sub_address_keys(
                sd.view_sec, sd.spend_pub, major, minor
            )
            addr = encode_addr(
                net_version(net_type, is_subaddr=True),
                crypto.encodepoint(D),
                crypto.encodepoint(C),
            )
            yield (major, minor, addr)


def main(args):
    loop = asyncio.get_event_loop()
    loop.run_until_complete(amain(args))
    loop.close()


if __name__ == "__main__":
    main(args)



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class DebugLinkMemoryRead(p.MessageType):
    MESSAGE_WIRE_TYPE = 110
    FIELDS = {
        1: ('address', p.UVarintType, 0),
        2: ('length', p.UVarintType, 0),
    }

    def __init__(
        self,
        address: int = None,
        length: int = None,
    ) -> None:
        self.address = address
        self.length = length



from monero_serialize import xmrserialize
from monero_serialize.xmrtypes import RctType, ECKey, EcdhInfo, KeyV, RctSigBase

from monero_glue.xmr import crypto
from monero_glue.xmr.sub.keccak_hasher import HashWrapper


class PreMlsagHasher(object):
    """
    Iterative construction of the pre_mlsag_hash
    """

    def __init__(self, state=None):
        from monero_glue.xmr.sub.keccak_hasher import HashWrapper, KeccakArchive

        self.is_simple = state[0] if state else None
        self.state = state[1] if state else 0
        self.kc_master = HashWrapper(state[2] if state else crypto.get_keccak())
        self.rsig_hasher = state[3] if state else crypto.get_keccak()
        self.rtcsig_hasher = None
        if state:
            self.rtcsig_hasher = KeccakArchive(state[4]) if state[4] else None
        else:
            self.rtcsig_hasher = KeccakArchive()

    def state_save(self):
        return (
            self.is_simple,
            self.state,
            self.kc_master.ctx,
            self.rsig_hasher,
            self.rtcsig_hasher.ctx() if self.rtcsig_hasher else None,
        )

    def state_load(self, x):
        from monero_glue.xmr.sub.keccak_hasher import HashWrapper, KeccakArchive

        self.is_simple = x[0]
        self.state = x[1]
        self.kc_master = HashWrapper(x[2])
        self.rsig_hasher = x[3]
        if x[4]:
            self.rtcsig_hasher = KeccakArchive(x[4])
        else:
            self.rtcsig_hasher = None

    def init(self, is_simple):
        if self.state != 0:
            raise ValueError("State error")

        self.state = 1
        self.is_simple = is_simple

    async def set_message(self, message):
        self.kc_master.update(message)

    async def set_type_fee(self, rv_type, fee):
        if self.state != 1:
            raise ValueError("State error")
        self.state = 2

        rfields = RctSigBase.f_specs()
        await self.rtcsig_hasher.ar.message_field(
            None, field=rfields[0], fvalue=rv_type
        )
        await self.rtcsig_hasher.ar.message_field(None, field=rfields[1], fvalue=fee)

    async def set_pseudo_out(self, out):
        if self.state != 2 and self.state != 3:
            raise ValueError("State error")
        self.state = 3

        await self.rtcsig_hasher.ar.field(out, KeyV.ELEM_TYPE)

    async def set_ecdh(self, ecdh):
        if self.state != 2 and self.state != 3 and self.state != 4:
            raise ValueError("State error")
        self.state = 4

        await self.rtcsig_hasher.ar.field(ecdh, EcdhInfo.ELEM_TYPE)

    async def set_out_pk(self, out_pk, mask=None):
        if self.state != 4 and self.state != 5:
            raise ValueError("State error")
        self.state = 5

        await self.rtcsig_hasher.ar.field(mask if mask else out_pk.mask, ECKey)

    async def rctsig_base_done(self):
        if self.state != 5:
            raise ValueError("State error")
        self.state = 6

        c_hash = self.rtcsig_hasher.kwriter.get_digest()
        self.kc_master.update(c_hash)
        self.rtcsig_hasher = None

    async def rsig_val(self, p, bulletproof, raw=False):
        if self.state == 8:
            raise ValueError("State error")

        if raw:
            self.rsig_hasher.update(p)
            return

        if bulletproof:
            self.rsig_hasher.update(p.A)
            self.rsig_hasher.update(p.S)
            self.rsig_hasher.update(p.T1)
            self.rsig_hasher.update(p.T2)
            self.rsig_hasher.update(p.taux)
            self.rsig_hasher.update(p.mu)
            for i in range(len(p.L)):
                self.rsig_hasher.update(p.L[i])
            for i in range(len(p.R)):
                self.rsig_hasher.update(p.R[i])
            self.rsig_hasher.update(p.a)
            self.rsig_hasher.update(p.b)
            self.rsig_hasher.update(p.t)

        else:
            for i in range(64):
                self.rsig_hasher.update(p.asig.s0[i])
            for i in range(64):
                self.rsig_hasher.update(p.asig.s1[i])
            self.rsig_hasher.update(p.asig.ee)
            for i in range(64):
                self.rsig_hasher.update(p.Ci[i])

    async def get_digest(self):
        if self.state != 6:
            raise ValueError("State error")
        self.state = 8

        c_hash = self.rsig_hasher.digest()
        self.rsig_hasher = None

        self.kc_master.update(c_hash)
        return self.kc_master.digest()


async def get_pre_mlsag_hash(rv):
    """
    Generates final message for the Ring CT signature

    :param rv:
    :type rv: RctSig
    :return:
    """
    from monero_glue.xmr.sub.keccak_hasher import get_keccak_writer

    kc_master = HashWrapper(crypto.get_keccak())
    kc_master.update(rv.message)

    is_simple = rv.type in [RctType.Simple, RctType.SimpleBulletproof]
    inputs = len(rv.pseudoOuts) if is_simple else 0
    outputs = len(rv.ecdhInfo)

    kwriter = get_keccak_writer()
    ar = xmrserialize.Archive(kwriter, True)
    await rv.serialize_rctsig_base(ar, inputs, outputs)
    c_hash = kwriter.get_digest()
    kc_master.update(c_hash)

    kc = crypto.get_keccak()
    if rv.type in [RctType.FullBulletproof, RctType.SimpleBulletproof]:
        for p in rv.p.bulletproofs:
            kc.update(p.A)
            kc.update(p.S)
            kc.update(p.T1)
            kc.update(p.T2)
            kc.update(p.taux)
            kc.update(p.mu)
            for i in range(len(p.L)):
                kc.update(p.L[i])
            for i in range(len(p.R)):
                kc.update(p.R[i])
            kc.update(p.a)
            kc.update(p.b)
            kc.update(p.t)

    else:
        for r in rv.p.rangeSigs:
            for i in range(64):
                kc.update(r.asig.s0[i])
            for i in range(64):
                kc.update(r.asig.s1[i])
            kc.update(r.asig.ee)
            for i in range(64):
                kc.update(r.Ci[i])

    c_hash = kc.digest()
    kc_master.update(c_hash)
    return kc_master.digest()



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class Features(p.MessageType):
    MESSAGE_WIRE_TYPE = 17
    FIELDS = {
        1: ('vendor', p.UnicodeType, 0),
        2: ('major_version', p.UVarintType, 0),
        3: ('minor_version', p.UVarintType, 0),
        4: ('patch_version', p.UVarintType, 0),
        5: ('bootloader_mode', p.BoolType, 0),
        6: ('device_id', p.UnicodeType, 0),
        7: ('pin_protection', p.BoolType, 0),
        8: ('passphrase_protection', p.BoolType, 0),
        9: ('language', p.UnicodeType, 0),
        10: ('label', p.UnicodeType, 0),
        12: ('initialized', p.BoolType, 0),
        13: ('revision', p.BytesType, 0),
        14: ('bootloader_hash', p.BytesType, 0),
        15: ('imported', p.BoolType, 0),
        16: ('pin_cached', p.BoolType, 0),
        17: ('passphrase_cached', p.BoolType, 0),
        18: ('firmware_present', p.BoolType, 0),
        19: ('needs_backup', p.BoolType, 0),
        20: ('flags', p.UVarintType, 0),
        21: ('model', p.UnicodeType, 0),
        22: ('fw_major', p.UVarintType, 0),
        23: ('fw_minor', p.UVarintType, 0),
        24: ('fw_patch', p.UVarintType, 0),
        25: ('fw_vendor', p.UnicodeType, 0),
        26: ('fw_vendor_keys', p.BytesType, 0),
        27: ('unfinished_backup', p.BoolType, 0),
    }

    def __init__(
        self,
        vendor: str = None,
        major_version: int = None,
        minor_version: int = None,
        patch_version: int = None,
        bootloader_mode: bool = None,
        device_id: str = None,
        pin_protection: bool = None,
        passphrase_protection: bool = None,
        language: str = None,
        label: str = None,
        initialized: bool = None,
        revision: bytes = None,
        bootloader_hash: bytes = None,
        imported: bool = None,
        pin_cached: bool = None,
        passphrase_cached: bool = None,
        firmware_present: bool = None,
        needs_backup: bool = None,
        flags: int = None,
        model: str = None,
        fw_major: int = None,
        fw_minor: int = None,
        fw_patch: int = None,
        fw_vendor: str = None,
        fw_vendor_keys: bytes = None,
        unfinished_backup: bool = None,
    ) -> None:
        self.vendor = vendor
        self.major_version = major_version
        self.minor_version = minor_version
        self.patch_version = patch_version
        self.bootloader_mode = bootloader_mode
        self.device_id = device_id
        self.pin_protection = pin_protection
        self.passphrase_protection = passphrase_protection
        self.language = language
        self.label = label
        self.initialized = initialized
        self.revision = revision
        self.bootloader_hash = bootloader_hash
        self.imported = imported
        self.pin_cached = pin_cached
        self.passphrase_cached = passphrase_cached
        self.firmware_present = firmware_present
        self.needs_backup = needs_backup
        self.flags = flags
        self.model = model
        self.fw_major = fw_major
        self.fw_minor = fw_minor
        self.fw_patch = fw_patch
        self.fw_vendor = fw_vendor
        self.fw_vendor_keys = fw_vendor_keys
        self.unfinished_backup = unfinished_backup



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class MoneroTransactionSetInputAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 503
    FIELDS = {
        1: ('vini', p.BytesType, 0),
        2: ('vini_hmac', p.BytesType, 0),
        3: ('pseudo_out', p.BytesType, 0),
        4: ('pseudo_out_hmac', p.BytesType, 0),
        5: ('alpha_enc', p.BytesType, 0),
        6: ('spend_enc', p.BytesType, 0),
    }

    def __init__(
        self,
        vini: bytes = None,
        vini_hmac: bytes = None,
        pseudo_out: bytes = None,
        pseudo_out_hmac: bytes = None,
        alpha_enc: bytes = None,
        spend_enc: bytes = None,
    ) -> None:
        self.vini = vini
        self.vini_hmac = vini_hmac
        self.pseudo_out = pseudo_out
        self.pseudo_out_hmac = pseudo_out_hmac
        self.alpha_enc = alpha_enc
        self.spend_enc = spend_enc



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p
from .HDNodeType import HDNodeType


class CardanoPublicKey(p.MessageType):
    MESSAGE_WIRE_TYPE = 306
    FIELDS = {
        1: ('xpub', p.UnicodeType, 0),
        2: ('node', HDNodeType, 0),
        3: ('root_hd_passphrase', p.UnicodeType, 0),
    }

    def __init__(
        self,
        xpub: str = None,
        node: HDNodeType = None,
        root_hd_passphrase: str = None,
    ) -> None:
        self.xpub = xpub
        self.node = node
        self.root_hd_passphrase = root_hd_passphrase



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class WordAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 47
    FIELDS = {
        1: ('word', p.UnicodeType, 0),  # required
    }

    def __init__(
        self,
        word: str = None,
    ) -> None:
        self.word = word



from monero_serialize.core.base_types import BoolType, UVarintType
from monero_serialize.core.message_types import BlobType, ContainerType, MessageType
from monero_serialize.xmrtypes import TxDestinationEntry, SecretKey


class TsxData(MessageType):
    """
    TsxData, initial input to the transaction processing.
    Serialization structure for easy hashing.
    """

    __slots__ = (
        "version",
        "payment_id",
        "unlock_time",
        "outputs",
        "change_dts",
        "num_inputs",
        "mixin",
        "fee",
        "account",
        "minor_indices",
        "is_multisig",
        "exp_tx_prefix_hash",
        "use_tx_keys",
        "is_bulletproof",
    )

    @classmethod
    def f_specs(cls):
        return (
            ("version", UVarintType),
            ("payment_id", BlobType),
            ("unlock_time", UVarintType),
            ("outputs", ContainerType, TxDestinationEntry),
            ("change_dts", TxDestinationEntry),
            ("num_inputs", UVarintType),
            ("mixin", UVarintType),
            ("fee", UVarintType),
            ("account", UVarintType),
            ("minor_indices", ContainerType, UVarintType),
            ("is_multisig", BoolType),
            ("exp_tx_prefix_hash", BlobType),  # expected prefix hash, bail on error
            ("use_tx_keys", ContainerType, SecretKey),  # use this secret key, multisig
            ("is_bulletproof", BoolType),
        )

    def __init__(self, payment_id=None, outputs=None, change_dts=None, **kwargs):
        super().__init__(**kwargs)

        self.payment_id = payment_id
        self.change_dts = change_dts
        self.fee = 0
        self.account = 0
        self.minor_indices = [0]
        self.outputs = outputs if outputs else []  # type: list[TxDestinationEntry]
        self.is_multisig = False
        self.is_bulletproof = False
        self.exp_tx_prefix_hash = b""
        self.use_tx_keys = []



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class MoneroRingCtSig(p.MessageType):
    FIELDS = {
        1: ('txn_fee', p.UVarintType, 0),
        2: ('message', p.BytesType, 0),
        3: ('rv_type', p.UVarintType, 0),
    }

    def __init__(
        self,
        txn_fee: int = None,
        message: bytes = None,
        rv_type: int = None,
    ) -> None:
        self.txn_fee = txn_fee
        self.message = message
        self.rv_type = rv_type



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class WordRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 46
    FIELDS = {
        1: ('type', p.UVarintType, 0),
    }

    def __init__(
        self,
        type: int = None,
    ) -> None:
        self.type = type



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class TezosOriginationType(p.MessageType):
    FIELDS = {
        1: ('manager_pubkey', p.BytesType, 0),
        2: ('balance', p.UVarintType, 0),
        3: ('spendable', p.BoolType, 0),
        4: ('delegatable', p.BoolType, 0),
        5: ('delegate', p.BytesType, 0),
        6: ('script', p.BytesType, 0),
    }

    def __init__(
        self,
        manager_pubkey: bytes = None,
        balance: int = None,
        spendable: bool = None,
        delegatable: bool = None,
        delegate: bytes = None,
        script: bytes = None,
    ) -> None:
        self.manager_pubkey = manager_pubkey
        self.balance = balance
        self.spendable = spendable
        self.delegatable = delegatable
        self.delegate = delegate
        self.script = script



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

from monero_glue import protobuf
from monero_glue.compat import gc
from monero_glue.messages import (
    MoneroAccountPublicAddress,
    MoneroTransactionData,
    MoneroTransactionDestinationEntry,
)
from monero_glue.xmr import crypto
from monero_glue.xmr.tsx_data import TsxData
from monero_serialize import xmrserialize, xmrtypes


class TrezorError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for kw in kwargs:
            setattr(self, kw, kwargs[kw])


class TrezorSecurityError(TrezorError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TrezorInvalidStateError(TrezorError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TrezorTxPrefixHashNotMatchingError(TrezorError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class StdObj(object):
    def __init__(self, **kwargs):
        for kw in kwargs:
            setattr(self, kw, kwargs[kw])


def compute_tx_key(spend_key_private, tx_prefix_hash, salt=None, rand_mult=None):
    """

    :param spend_key_private:
    :param tx_prefix_hash:
    :param salt:
    :param rand_mult:
    :return:
    """
    if not salt:
        salt = crypto.random_bytes(32)

    if not rand_mult:
        rand_mult_num = crypto.random_scalar()
        rand_mult = crypto.encodeint(rand_mult_num)
    else:
        rand_mult_num = crypto.decodeint(rand_mult)

    rand_inp = crypto.sc_add(spend_key_private, rand_mult_num)
    passwd = crypto.keccak_2hash(crypto.encodeint(rand_inp) + tx_prefix_hash)
    tx_key = crypto.compute_hmac(salt, passwd)
    # tx_key = crypto.pbkdf2(passwd, salt, count=100)
    return tx_key, salt, rand_mult


def translate_monero_dest_entry(dst_entry: MoneroTransactionDestinationEntry):
    d = xmrtypes.TxDestinationEntry()
    d.amount = dst_entry.amount
    d.is_subaddress = dst_entry.is_subaddress
    d.addr = xmrtypes.AccountPublicAddress(
        m_spend_public_key=dst_entry.addr.spend_public_key,
        m_view_public_key=dst_entry.addr.view_public_key,
    )
    return d


def translate_monero_dest_entry_pb(dst_entry: xmrtypes.TxDestinationEntry):
    d = MoneroTransactionDestinationEntry(
        amount=dst_entry.amount,
        is_subaddress=dst_entry.is_subaddress,
        addr=MoneroAccountPublicAddress(
            spend_public_key=dst_entry.addr.m_spend_public_key,
            view_public_key=dst_entry.addr.m_view_public_key,
        ),
    )
    return d


async def translate_tsx_data(tsx_data: MoneroTransactionData):
    tsxd = TsxData()
    for fld in TsxData.f_specs():
        fname = fld[0]
        if hasattr(tsx_data, fname):
            setattr(tsxd, fname, getattr(tsx_data, fname))

    if tsx_data.change_dts:
        tsxd.change_dts = translate_monero_dest_entry(tsx_data.change_dts)

    tsxd.outputs = [translate_monero_dest_entry(x) for x in tsx_data.outputs]
    return tsxd


async def translate_tsx_data_pb(tsx_data: TsxData):
    tsxd = MoneroTransactionData()
    for fld in TsxData.f_specs():
        fname = fld[0]
        if hasattr(tsx_data, fname):
            setattr(tsxd, fname, getattr(tsx_data, fname))

    if tsx_data.change_dts:
        tsxd.change_dts = translate_monero_dest_entry_pb(tsx_data.change_dts)

    tsxd.outputs = [translate_monero_dest_entry_pb(x) for x in tsx_data.outputs]
    return tsxd


async def parse_msg(bts, msg):
    reader = xmrserialize.MemoryReaderWriter(bytearray(bts))
    ar = xmrserialize.Archive(reader, False)
    return await ar.message(msg)


async def parse_pb_msg(bts, msg):
    reader = xmrserialize.MemoryReaderWriter(bytearray(bts))
    return await protobuf.load_message(reader, msg)


async def parse_src_entry(bts):
    return await parse_msg(bts, xmrtypes.TxSourceEntry())


async def parse_dst_entry(bts):
    return await parse_msg(bts, xmrtypes.TxDestinationEntry())


async def parse_vini(bts):
    return await parse_msg(bts, xmrtypes.TxinToKey())


async def dump_msg(msg, preallocate=None, msg_type=None):
    writer = xmrserialize.MemoryReaderWriter(preallocate=preallocate)
    ar = xmrserialize.Archive(writer, True)
    await ar.message(msg, msg_type=msg_type)
    return writer.get_buffer()


async def dump_pb_msg(msg):
    writer = xmrserialize.MemoryReaderWriter()
    await protobuf.dump_message(writer, msg)
    return bytes(writer.get_buffer())


async def dump_msg_gc(msg, preallocate=None, msg_type=None, del_msg=False):
    b = await dump_msg(msg, preallocate=preallocate, msg_type=msg_type)
    if del_msg:
        del msg

    import gc

    gc.collect()
    return b



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

#
# Note pickling is used for message serialization.
# This is just for the prototyping & fast PoC, pickling wont be used in the production.
# Instead, protobuf messages will be defined and parsed to avoid malicious pickling.
#

import binascii
import logging
import pickle

import requests
from monero_glue import protobuf
from monero_glue.hwtoken import token
from monero_glue.protocol_base import messages
from monero_serialize import xmrserialize

logger = logging.getLogger(__name__)


class TokenProxy(token.TokenLite):
    """
    Trezor proxy calls to the remote server
    """

    def __init__(self, url=None, *args, **kwargs):
        super().__init__()
        self.url = "http://127.0.0.1:46123" if url is None else url
        self.endpoint = "%s/api/v1.0" % self.url

    async def transfer(self, method, cmd, payload):
        endp = "%s/%s" % (self.endpoint, method)
        req = {"cmd": cmd, "payload": payload}
        resp = requests.post(endp, json=req)
        resp.raise_for_status()
        return resp.json()

    async def transfer_pickle(self, method, action, *args, **kwargs):
        logger.debug("Action: %s" % action)
        to_pickle = (args, kwargs)
        pickled_data = pickle.dumps(to_pickle)
        payload = binascii.hexlify(pickled_data).decode("utf8")

        resp = await self.transfer(method, action, payload)
        pickle_data = binascii.unhexlify(resp["payload"].encode("utf8"))
        logger.debug(
            "Req size: %s, response size: %s" % (len(pickled_data), len(pickle_data))
        )

        res = pickle.loads(pickle_data)
        return res

    async def transfer_protobuf(self, method, msg: protobuf.MessageType):
        logger.debug("Method: %s" % method)
        writer = xmrserialize.MemoryReaderWriter()

        await protobuf.dump_message(writer, msg)
        proto_bin = bytes(writer.get_buffer())
        payload = {
            "msg_type": messages.get_message_type(msg),
            "msg": binascii.hexlify(proto_bin).decode("utf8"),
        }

        resp = await self.transfer(method, "", payload)
        resp_bin = binascii.unhexlify(resp["payload"]["msg"].encode("utf8"))
        logger.debug(
            "Req size: %s, response size: %s" % (len(proto_bin), len(resp_bin))
        )

        reader = xmrserialize.MemoryReaderWriter(bytearray(resp_bin))
        res = await protobuf.load_message(
            reader, messages.get_message_from_type(resp["payload"]["msg_type"])
        )
        return res

    async def call(self, msg, recode=True):
        return await self.transfer_protobuf("call", msg)

    async def ping(self, message=None, **kwargs):
        resp = requests.get("%s/ping" % self.endpoint)
        resp.raise_for_status()
        return resp.json()

    async def get_view_key(self, msg):
        return await self.transfer_protobuf("watch_only", msg)

    async def tsx_sign(self, msg):
        return await self.transfer_protobuf("tx_sign", msg)

    async def key_image_sync(self, msg, *args, **kwargs):
        return await self.transfer_protobuf("ki_sync", msg)



'''
Extremely minimal streaming codec for a subset of protobuf.  Supports uint32,
bytes, string, embedded message and repeated fields.

For de-serializing (loading) protobuf types, object with `AsyncReader`
interface is required:

>>> class AsyncReader:
>>>     async def areadinto(self, buffer):
>>>         """
>>>         Reads `len(buffer)` bytes into `buffer`, or raises `EOFError`.
>>>         """

For serializing (dumping) protobuf types, object with `AsyncWriter` interface is
required:

>>> class AsyncWriter:
>>>     async def awrite(self, buffer):
>>>         """
>>>         Writes all bytes from `buffer`, or raises `EOFError`.
>>>         """
'''

# from micropython import const


def const(x):
    return x


_UVARINT_BUFFER = bytearray(1)


async def load_uvarint(reader):
    buffer = _UVARINT_BUFFER
    result = 0
    shift = 0
    byte = 0x80
    while byte & 0x80:
        await reader.areadinto(buffer)
        byte = buffer[0]
        result += (byte & 0x7F) << shift
        shift += 7
    return result


async def dump_uvarint(writer, n):
    if n < 0:
        raise ValueError("Cannot dump signed value, convert it to unsigned first.")
    buffer = _UVARINT_BUFFER
    shifted = True
    while shifted:
        shifted = n >> 7
        buffer[0] = (n & 0x7F) | (0x80 if shifted else 0x00)
        await writer.awrite(buffer)
        n = shifted


# protobuf interleaved signed encoding:
# https://developers.google.com/protocol-buffers/docs/encoding#structure
# the idea is to save the sign in LSbit instead of twos-complement.
# so counting up, you go: 0, -1, 1, -2, 2, ... (as the first bit changes, sign flips)
#
# To achieve this with a twos-complement number:
# 1. shift left by 1, leaving LSbit free
# 2. if the number is negative, do bitwise negation.
#    This keeps positive number the same, and converts negative from twos-complement
#    to the appropriate value, while setting the sign bit.
#
# The original algorithm makes use of the fact that arithmetic (signed) shift
# keeps the sign bits, so for a n-bit number, (x >> n) gets us "all sign bits".
# Then you can take "number XOR all-sign-bits", which is XOR 0 (identity) for positive
# and XOR 1 (bitwise negation) for negative. Cute and efficient.
#
# But this is harder in Python because we don't natively know the bit size of the number.
# So we have to branch on whether the number is negative.


def sint_to_uint(sint):
    res = sint << 1
    if sint < 0:
        res = ~res
    return res


def uint_to_sint(uint):
    sign = uint & 1
    res = uint >> 1
    if sign:
        res = ~res
    return res


class UVarintType:
    WIRE_TYPE = 0


class SVarintType:
    WIRE_TYPE = 0


class BoolType:
    WIRE_TYPE = 0


class BytesType:
    WIRE_TYPE = 2


class UnicodeType:
    WIRE_TYPE = 2


class MessageType:
    WIRE_TYPE = 2
    FIELDS = {}

    def __init__(self, **kwargs):
        for kw in kwargs:
            setattr(self, kw, kwargs[kw])

    def __eq__(self, rhs):
        return self.__class__ is rhs.__class__ and self.__dict__ == rhs.__dict__

    def __repr__(self):
        return "<%s>" % self.__class__.__name__


class LimitedReader:
    def __init__(self, reader, limit):
        self.reader = reader
        self.limit = limit

    async def areadinto(self, buf):
        if self.limit < len(buf):
            raise EOFError
        else:
            nread = await self.reader.areadinto(buf)
            self.limit -= nread
            return nread


class CountingWriter:
    def __init__(self):
        self.size = 0

    async def awrite(self, buf):
        nwritten = len(buf)
        self.size += nwritten
        return nwritten


FLAG_REPEATED = const(1)


async def load_message(reader, msg_type):
    fields = msg_type.FIELDS
    msg = msg_type()

    while True:
        try:
            fkey = await load_uvarint(reader)
        except EOFError:
            break  # no more fields to load

        ftag = fkey >> 3
        wtype = fkey & 7

        field = fields.get(ftag, None)

        if field is None:  # unknown field, skip it
            if wtype == 0:
                await load_uvarint(reader)
            elif wtype == 2:
                ivalue = await load_uvarint(reader)
                await reader.areadinto(bytearray(ivalue))
            else:
                raise ValueError
            continue

        fname, ftype, fflags = field
        if wtype != ftype.WIRE_TYPE:
            raise TypeError  # parsed wire type differs from the schema

        ivalue = await load_uvarint(reader)

        if ftype is UVarintType:
            fvalue = ivalue
        elif ftype is SVarintType:
            fvalue = uint_to_sint(ivalue)
        elif ftype is BoolType:
            fvalue = bool(ivalue)
        elif ftype is BytesType:
            fvalue = bytearray(ivalue)
            await reader.areadinto(fvalue)
        elif ftype is UnicodeType:
            fvalue = bytearray(ivalue)
            await reader.areadinto(fvalue)
            fvalue = str(fvalue, "utf8")
        elif issubclass(ftype, MessageType):
            fvalue = await load_message(LimitedReader(reader, ivalue), ftype)
        else:
            raise TypeError  # field type is unknown

        if fflags & FLAG_REPEATED:
            pvalue = getattr(msg, fname, [])
            pvalue.append(fvalue)
            fvalue = pvalue
        setattr(msg, fname, fvalue)

    # fill missing fields
    for tag in msg.FIELDS:
        field = msg.FIELDS[tag]
        if not hasattr(msg, field[0]):
            setattr(msg, field[0], None)

    return msg


async def dump_message(writer, msg):
    repvalue = [0]
    mtype = msg.__class__
    fields = mtype.FIELDS

    for ftag in fields:
        fname, ftype, fflags = fields[ftag]

        fvalue = getattr(msg, fname, None)
        if fvalue is None:
            continue

        fkey = (ftag << 3) | ftype.WIRE_TYPE

        if not fflags & FLAG_REPEATED:
            repvalue[0] = fvalue
            fvalue = repvalue

        for svalue in fvalue:
            await dump_uvarint(writer, fkey)

            if ftype is UVarintType:
                await dump_uvarint(writer, svalue)

            elif ftype is SVarintType:
                await dump_uvarint(writer, sint_to_uint(svalue))

            elif ftype is BoolType:
                await dump_uvarint(writer, int(svalue))

            elif ftype is BytesType:
                await dump_uvarint(writer, len(svalue))
                await writer.awrite(svalue)

            elif ftype is UnicodeType:
                bvalue = bytes(svalue, "utf8")
                await dump_uvarint(writer, len(bvalue))
                await writer.awrite(bvalue)

            elif issubclass(ftype, MessageType):
                counter = CountingWriter()
                await dump_message(counter, svalue)
                await dump_uvarint(writer, counter.size)
                await dump_message(writer, svalue)

            else:
                raise TypeError



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class CardanoTxAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 309
    FIELDS = {
        1: ('transaction', p.BytesType, 0),
    }

    def __init__(
        self,
        transaction: bytes = None,
    ) -> None:
        self.transaction = transaction
